/* Jon Arney, (c) 2000 */
/* Ashley Jennings (c) 2005-2011 gumnut GUI code */


/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>

/* 0.4.28.c17 */
#include "sh_sys_types.h"

#include <sys/stat.h>

#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <stdarg.h>

#include "cache.h"
#include "conf.h"
#include "connection.h"
#include "hash.h"
#include "host.h"
#include "http.h"
#include "lib.h"
#include "net.h"
#include "qry.h"
#include "transfer.h"
#include "share.h"
#include "threads.h"
#include "ui.h"
#include "cli.h"
//#include "hashlib/hashlib.h"

#define GNUT_HTTP_CACHE_EXPIRE 20

#define PRINTFDBG 0

/* debug helper macros to replace typing printf statements */
#define PI(i) if (PRINTFDBG) { printf( "+++  [%d] [ " #i " ] [ %d ]  --  [line #] [int identifier] [value]\n",__LINE__, i ); }
#define PS(s) if (PRINTFDBG) { printf( "+++  [%d] [ " #s " ] [ %s ]  --  [line #] [str identifier] [value]\n",__LINE__, s ); }

int guidispindex = 1; /* global var to hold value of which query to show contents of in the HTTP GUI */

uint32 urate_history[10];
uint32 best_urate;
int urate_clipped;
int urate_measured;
int gh_did_upload;

#ifdef GNUT_HTTP_SEARCH
int cgi_parse( char * CGI, char * field, char * value );
static void gnut_guid_to_string(char *string, char *guid);
/* static void gnut_string_to_guid(char *guid, char *string); */

Gnut_Hash *query_cache;

char *html_template_head;
char *html_template_foot;

void fre_ghq(gnut_http_query_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

/* Argument parsing routines
 * Return: 1 if invalid CGI string
 *         0 if   valid CGI string   */
int cgi_parse(char * CGI, char * field, char * value)
{
  char  * start;
  char  * end;
  char  * eq;
  char  tmp[8192];
  char  tmp2[8192];
  int  flag=1;

  if (! strstr(CGI, "=")) {
    /* If no equal sign then this is an invalid CGI string */
    GD_S(1, "This is _not_ a CGI string\n");
    return 1;
  }

  start = CGI;
  end   = start;

  while(*start &&  *end  && (flag == 1)) {
    /* Find the end of the CGI string */
    end = strstr( start, "&" );
    if (!end) {
      end = start + strlen(start);
    }

    /* Is this the correct one? */
    strncpy(tmp, start, end - start);
    tmp[end - start + 1] = '\0';
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, " xxxxa tmp= \n");
    GD_S(GNUT_HTTP_INTERNALS_DEBUG,tmp);

    GD_I(GNUT_HTTP_INTERNALS_DEBUG, strlen(tmp));
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, " xxxxa tmp=  yyyyy \n");

    cgi_to_normal(tmp2, tmp);

    if (! strcmp(tmp2, field) || ! strcmp(tmp, field)) {
      eq = strstr(start, "=");
      if (eq) {
	strncpy(tmp, eq + 1, end - eq + 1);
	tmp[ end - eq + 1 ] = '\0';

	cgi_to_normal(value, tmp);

	flag = 0;
      }
    }

    /* Move on to next one */
    start = end + 1;
  }

  return flag;
}

void cgi_to_normal(char * news, char * olds)
{
  int  done = 0;
  int  c;
  char pbuf[3];

  while(!done) {
    if (G_ISALNUM(* olds) || * olds == '_' || * olds == '.' || * olds == '-' ) {
      *news = *olds;
    } else if (*olds == '+') {
      *news = ' ';
    } else if (*olds == '%') {
      olds++;
      pbuf[0] = *olds;
      pbuf[1] = *(olds + 1);
      pbuf[2] = '\0';
      sscanf(pbuf, "%02x", &c);
      * news = c;
      olds++;
    } else {
      * news = '\0';
      done = 1;
    }
    olds++;
    news++;
  }

  return;
}

static uchar gnut_http_query_cache_hash(void *p_data)
{
  gnut_http_query_t *data;
  uchar hashval;
  int i;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cache_hash IN\n");

  data = (gnut_http_query_t *)p_data;

  hashval = 0;
  for (i = 1; i < 16; i++) {
    if (i != 6) {
      hashval += data->guid[i];
    }
  }
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cache_hash OUT\n");
  return hashval;
}

static int gnut_http_query_cache_compare(void *p_a, void *p_b)
{
  gnut_http_query_t *a;
  gnut_http_query_t *b;
  char stringa[65];
  char stringb[65];

  a = (gnut_http_query_t *)p_a;
  b = (gnut_http_query_t *)p_b;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cache_compare IN\n");

  /*
   * They match if the guid's match.
   */
  gnut_guid_to_string(stringa, a->guid);
  gnut_guid_to_string(stringb, b->guid);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Hasha = ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, stringa);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Hashb = ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, stringb);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");
  a->guid[6] = '\0';
  b->guid[6] = '\0';
  if (memcmp(a->guid, b->guid, sizeof(a->guid)) == 0) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cache_compare OUT - Found an equal hash by GUID\n");
    return 0;
  }

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cache_compare OUT - Hash not equal\n");
  return -1;
}

static void gnut_http_query_cleanup(void)
{
  gnut_http_query_t *req;
  Gnut_List *list;
  int i;
  time_t last_time_to_keep;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cleanup IN\n");

  if (!query_cache) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cleanup OUT - nothing to clean\n");
    return;
  }

  /* Do garbage collection on the hash.
   * based on 'oldest'.  Anything older than
   * EXPIRE minutes gets deleted. */
  last_time_to_keep = time(0);
  last_time_to_keep -= (GNUT_HTTP_CACHE_EXPIRE*60);

  /* For each element of the hash,
   * remove the oldest. */
  for (i = 0; i < 256; i++) {
    list = query_cache->list[i];

    /* Walk the list looking for
     * old guys, and delete them. */
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Starting garbage collection with ");
    GD_I(GNUT_HTTP_INTERNALS_DEBUG, gnut_list_size(query_cache->list[i]));
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

    while (1) {
      if (!list) {
	break;
      }
      req = list->data;
      if (!req) {
	break;
      }

      /* This is a hack because I'm trying to remove an entry from the same
       * list I'm walking. */
      if (req->last_used < last_time_to_keep) {
	GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Removing an old req from the list\n");
	list = gnut_list_remove(list, req, 4);
	GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Fr""eeing an old one\n");
	fre_ghq(&req, 383);
	GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Re-assigning the list\n");
	query_cache->list[i] = list;
      } else {
	list = gnut_list_next(list);
      }
    }
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Finished garbage collection, now have ");
    GD_I(GNUT_HTTP_INTERNALS_DEBUG, gnut_list_size(query_cache->list[i]));
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");
  }
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_cleanup OUT\n");
}

Gnut_List *gnut_http_query_get(gnut_transfer *gt, char *guid)
{
  gnut_http_query_t *req, req_findme;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_get IN\n");

  memcpy(req_findme.guid, guid, 16);
  if (!query_cache) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_get OUT - no cache yet\n");
    return 0;
  }
  dqi(0x0157);
  req = gnut_hash_find(query_cache, &req_findme);
  if (!req) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_get OUT - no entry found in cache\n");
    return 0;
  }

  req->last_used = time(0);

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_get OUT\n");


  int n;

   n = gnut_list_size(req->query_list);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "listlength   n=  ");
     GD_I(GNUT_HTTP_INTERNALS_DEBUG, n);

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "ggggg gnut_http_query_get debugfoo OUT\n");


//  qsort((query_resp *)req->query_list, n, sizeof(query_resp *),
//          query_response_compare);



  return req->query_list;
}

void gnut_http_query_new(gnut_transfer *gt, char *request, char *guid)
{
  gnut_http_query_t *req, req_findme;
  gnutella_packet *gpa;
  int i;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_new IN\n");

  if (request == 0) {
    /* The request was a frivolous one, ignore it. */
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_new OUT - the query was frivolous\n");
    return;
  }

  gnut_http_query_cleanup();

  /* Actually submit the request with the new MAC
   * and stuff. */
  gpa=gp_request_make(conf_get_str("mac"),request,conf_get_int("ttl"));
  for (i = 7; i < 16; i++) {
    gpa->gh[i] = guid[i];
  }
  send_to_all(gpa);

  if (!query_cache) {
    query_cache = gnut_hash_new(gnut_http_query_cache_hash,
        gnut_http_query_cache_compare);
  }

  /* See if we're already on the list.... */
  /*  memcpy(req_findme.ip, gt->ip, sizeof(gt->ip)); */
  /*  req_findme.port = gt->gt_port; */
  memcpy(req_findme.guid, gpa->gh, 16);
  dqi(0x0158);
  req = gnut_hash_find(query_cache, &req_findme);

  /* We're not on the list, so add us to the list. */
  if (!req) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Appending an http to the hash list\n");
    req = (gnut_http_query_t *)ymaloc(sizeof(gnut_http_query_t), 326);
    memcpy(req->guid, gpa->gh, 16);
    memcpy(req->ip, gt->ip, 4);
    req->port = gt->gt_port;
    req->last_used = time(0);
    req->query_list = 0;
    gnut_hash_insert(query_cache, req, 500);
  } else {
    memcpy(req->guid, gpa->gh, 16);
    req->last_used = time(0);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "Updating the GUID to the hash list\n");
  }
  fre_v(&(gpa->data), 183);
  fre_gpa(&gpa, 184);

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_query_new OUT\n");
}

/* Check a query reply (0x81) packet to see if it belongs to a pending
 * query issued through the HTTP interface, and if it is, append its
 * data to the list for that query. Returns 1 if this was done, 0 if not. */
int gnut_http_result_append(uint8 ip[4], uint8 port_le[2], uint32 ref,
			    uint32 speed, uint32 size, char *name,
			    uint8 guid[16], int16 htype, int16 rating)
{
  gnut_http_query_t *req, req_findme;
  query_resp *a_qrp;
  char string[64];

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_result_append IN\n");
  if (!query_cache) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_result_append OUT - no query to append result to\n");
    return 0;
  }

  memcpy(req_findme.guid, guid, sizeof(req_findme.guid));
  gnut_guid_to_string(string, guid);
  dqi(0x0159);
  req = gnut_hash_find(query_cache, &req_findme);
  if (!req) {
    /* It doesn't appear to be our packet. */
    GD_S(GNUT_HTTP_INTERNALS_DEBUG,
	 "gnut_http_result_append OUT - this query did not belong to us : ");
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, string);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");
    return 0;
  }
  a_qrp = query_create(327,  /* qr-new */
		       ip, port_le, ref, speed, size, guid,
		       htype, rating, 0, ' ', 0, 0, 0, name);

  req->last_used = time(0);
  req->query_list = gnut_list_prepend(req->query_list , a_qrp);

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "gnut_http_result_append OUT - this reply belonged to us : ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, string);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");
  return 1;
}

#define BUFSZ1 512  /* tagok */

int http_write_str(http_request_t *req, char *a)
{
  int ret;

  GD_S(HTTP_IO_DEBUG, "http_write_str IN ");
  GD_S(HTTP_IO_DEBUG, a);
  GD_S(HTTP_IO_DEBUG, "\n");

  if (!req) {
    return -1;
  }
  if (req->sock < 0) {
    return -1;
  }
  ret = write(req->sock, a, strlen(a));

  GD_S(HTTP_IO_DEBUG, "http_write_str OUT ");
     GD_I(HTTP_IO_DEBUG, ret);
  GD_S(HTTP_IO_DEBUG, "\n");
  return ret;

}

int http_write_va(http_request_t *req, int maxlen, char *format, ...)
{
  va_list argp;
  int ret;
  char *value;

  GD_S(HTTP_IO_DEBUG, "http_write_va IN\n");

  value = ymaloc(sizeof(char)*maxlen, 306);

  va_start(argp,format);
  vsprintf(value,format,argp);
  va_end(argp);

  ret = http_write_str(req, value);

  fre_str(&value, 379);

  GD_S(HTTP_IO_DEBUG, "http_write_va OUT ");
  GD_I(HTTP_IO_DEBUG, ret);
  GD_S(HTTP_IO_DEBUG, "\n");

  return ret;
}


/*
 *
 * ***  START modified http mini server with gumnut search capability  ***
 *
 */


/* gnut_http_file_create function to create a new leaf through the http gui */
int gnut_http_file_create(gnut_transfer *gt, http_request_t *req)
{
/*
  Gnut_List *list;
  query_resp *a_qrp;
  */
  char *create_string = 0;
  char *query_string;

  char leafname[256];

  /*
  char c_port[32];
  char c_address[BUFSZ1];
  unsigned int i_port;
  int n, i1;
  */
  int i;
  char guid[16];
  http_cookie_t *cookp;

  int ret2, ret3, ret4, ret5, ret6, ret7, ret8, ret9, ret10, ret11, ret12, ret13, ret14;
  char value[BUFSZ1] = "", ts[BUFSZ1] = "", pl[BUFSZ1] = "", cc[BUFSZ1] = "", pc[BUFSZ1] = "", to1[BUFSZ1] = "",to2[BUFSZ1] = "",to3[BUFSZ1] = "",to4[BUFSZ1] = "",to5[BUFSZ1] = "";
  char cn[2048] = "";

  char *tmp = 0;
  char *fpath = 0;
  FILE *file = 0;
  char filename[256];

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_file_create IN\n");

  /* Check to see if we got a GUID in a cookie already... */
  cookp = http_cookie_new();

  /* Create a new guid and set it as a cookie to the browser. */
  GD_S(1,"creating new cookie\n");

  cookp->value = ymaloc(sizeof(char)*33, 328);

  memcpy(guid, conf_get_str("mac"), 6);
  for (i=6; i<16; i++) {
    guid[i] = rand();
  }

  gnut_guid_to_string(cookp->value, guid);
  http_response_set_cookie(req, cookp);
  GD_S(GNUT_HTTP_DEBUG, "Setting the guid in a cookie ");
  GD_S(GNUT_HTTP_DEBUG, cookp->value);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  http_cookie_delete(cookp);

  query_string = 0;

  /* Get the CGI string out of the query. */
  query_string = strstr(req->url_file, "/create/?");
  if (query_string) {

    query_string += 9;

    ret3 = 0;

    ret2 = cgi_parse(query_string, "timestamps", ts);
    if (  !strcmp(ts, "") ) {
      ret3 = cgi_parse(query_string, "time_stamps", ts);
    }

    strncpy(pl,"earth",5);  /* static for now */

    ret4 = cgi_parse(query_string, "countrycodes", cc);
    if (  !strcmp(cc, "") ) {
      ret5 = cgi_parse(query_string, "recent_cc", cc);
    }

    ret6 = cgi_parse(query_string, "postcodes", pc);
    if (  !strcmp(pc, "") ) {
      ret7 = cgi_parse(query_string, "recent_pc", pc);
    }

    ret8 = cgi_parse(query_string, "topic1", to1);
    ret9 = cgi_parse(query_string, "topic2", to2);
    ret10 = cgi_parse(query_string, "topic3", to3);
    ret11 = cgi_parse(query_string, "topic4", to4);
    ret12 = cgi_parse(query_string, "topic5", to5);
    ret13 = cgi_parse(query_string, "planet", pl);
    ret14 = cgi_parse(query_string, "content", cn);


    GD_S(GNUT_HTTP_DEBUG, "ret2==  XxX");
    GD_I(GNUT_HTTP_DEBUG, ret2);
    GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");
    GD_S(GNUT_HTTP_DEBUG, "ret3==  XxX");
    GD_I(GNUT_HTTP_DEBUG, ret3);
    GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");

    // processing values
    create_string = value;

    if ( !strcmp(ts, "") | !strcmp(ts, "Current Time") ) {

      GD_S(GNUT_HTTP_DEBUG, " ts blank or CurTim in  ts==  XxX");
      GD_S(GNUT_HTTP_DEBUG, ts);
      GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");

      time_t tim;

      tim = time(0);
      strftime(ts,16,"%Y%m%d%H%M%S",localtime(&tim));

    GD_S(GNUT_HTTP_DEBUG, "YY1YYYYYYYYYYYYYYYYYYY==  XxX");
    GD_S(GNUT_HTTP_DEBUG, create_string);
    GD_S(GNUT_HTTP_DEBUG, "XxX\n");
    strcat(create_string,"leaf");
    }





    int j;
    int k;
    char nospace_topic1[256] = "";

    char topic_string[2048] = "";

    j = 0;
    k = 1;
    // parse the topic and convert each whitespace to a "-"
    for (i=0;i <= strlen(to1);i++) {

  GD_S(GNUT_HTTP_DEBUG, "11 iiiiiii=");
    GD_I(GNUT_HTTP_DEBUG, i);
  GD_S(GNUT_HTTP_DEBUG, "\n");
  GD_S(GNUT_HTTP_DEBUG, "for 1111 to1= ");
  GD_S(GNUT_HTTP_DEBUG, to1);
  GD_S(GNUT_HTTP_DEBUG, "\n");
  GD_S(GNUT_HTTP_DEBUG, "11 strlen(to1)=");
    GD_I(GNUT_HTTP_DEBUG, strlen(to1));
  GD_S(GNUT_HTTP_DEBUG, "\n");
  GD_S(GNUT_HTTP_DEBUG, "for 1111 nospace_topic1= ");
  GD_S(GNUT_HTTP_DEBUG, nospace_topic1);
  GD_S(GNUT_HTTP_DEBUG, "\n");


      if (i == strlen(to1)) {
  GD_S(GNUT_HTTP_DEBUG, "if222  \n");
        nospace_topic1[j] = atoi("\0");
        continue;
      }
      if (to1[i] == ' ' ) {
        if ( (to1[i + 1] == ' ') | (i == strlen(to1) - 2) ) {
          continue;
        }
        to1[i] = '-';
      }
      if (to1[i] == ',' ) {
        if (to1[i + 1] == ',' ) {
          continue;
	}
        nospace_topic1[j] = to1[1];
	j++;
        nospace_topic1[j] = atoi("t");
	j++;
        nospace_topic1[j] = atoi("O");
	j++;
        nospace_topic1[j] = k;
	j++;
	continue;
      }
      nospace_topic1[j] = to1[i];
      j++;
    }
    strcpy(to1,nospace_topic1);

    strcpy(topic_string,"_tO1");
    strcat(topic_string,to1);




    if ( strcmp(to2, "") ) {

  GD_S(GNUT_HTTP_DEBUG, "if22222\n");
    i = 0;
    j = 0;

    nospace_topic1[0] = '\0';

    for (i=0;i <= strlen(to2);i++) {

      if (i == strlen(to2)) {
        nospace_topic1[j] = atoi("\0");
        continue;
      }
      if (to2[i] == ' ' ) {
        if ( (to2[i + 1] == ' ') | (i == strlen(to2) - 2) ) {
          continue;
        }
        to2[i] = '-';
      }
      if (to2[i] == ',' ) {
        if (to2[i + 1] == ',' ) {
          continue;
	}
        nospace_topic1[j] = to2[1];
	j++;
        nospace_topic1[j] = atoi("t");
	j++;
        nospace_topic1[j] = atoi("O");
	j++;
        nospace_topic1[j] = k;
	j++;
	continue;
      }
      nospace_topic1[j] = to2[i];
      j++;
    }
    strcpy(to2,nospace_topic1);

    strcat(topic_string,"_tO2");
    strcat(topic_string,to2);
    }


    if ( strcmp(to3, "") ) {

  GD_S(GNUT_HTTP_DEBUG, "if33333\n");
    i = 0;
    j = 0;

    nospace_topic1[0] = '\0';

    for (i=0;i <= strlen(to3);i++) {

      if (i == strlen(to3)) {
        nospace_topic1[j] = atoi("\0");
        continue;
      }
      if (to3[i] == ' ' ) {
        if ( (to3[i + 1] == ' ') | (i == strlen(to3) - 2) ) {
          continue;
        }
        to3[i] = '-';
      }
      if (to3[i] == ',' ) {
        if (to3[i + 1] == ',' ) {
          continue;
	}
        nospace_topic1[j] = to3[1];
	j++;
        nospace_topic1[j] = atoi("t");
	j++;
        nospace_topic1[j] = atoi("O");
	j++;
        nospace_topic1[j] = k;
	j++;
	continue;
      }
      nospace_topic1[j] = to3[i];
      j++;
    }
    strcpy(to3,nospace_topic1);

    strcat(topic_string,"_tO3");
    strcat(topic_string,to3);
    }


    if ( strcmp(to4, "") ) {
  GD_S(GNUT_HTTP_DEBUG, "if44444\n");

    i = 0;
    j = 0;

    nospace_topic1[0] = '\0';

    for (i=0;i <= strlen(to4);i++) {

      if (i == strlen(to4)) {
        nospace_topic1[j] = atoi("\0");
        continue;
      }
      if (to4[i] == ' ' ) {
        if ( (to4[i + 1] == ' ') | (i == strlen(to4) - 2) ) {
          continue;
        }
        to4[i] = '-';
      }
      if (to4[i] == ',' ) {
        if (to4[i + 1] == ',' ) {
          continue;
	}
        nospace_topic1[j] = to4[1];
	j++;
        nospace_topic1[j] = atoi("t");
	j++;
        nospace_topic1[j] = atoi("O");
	j++;
        nospace_topic1[j] = k;
	j++;
	continue;
      }
      nospace_topic1[j] = to4[i];
      j++;
    }
    strcpy(to4,nospace_topic1);

    strcat(topic_string,"_tO4");
    strcat(topic_string,to4);
    }

    if ( strcmp(to5, "") ) {
  GD_S(GNUT_HTTP_DEBUG, "if5555 \n");

    i = 0;
    j = 0;

    nospace_topic1[0] = '\0';

    for (i=0;i <= strlen(to5);i++) {

      if (i == strlen(to5)) {
        nospace_topic1[j] = atoi("\0");
        continue;
      }
      if (to5[i] == ' ' ) {
        if ( (to5[i + 1] == ' ') | (i == strlen(to5) - 2) ) {
          continue;
        }
        to5[i] = '-';
      }
      if (to5[i] == ',' ) {
        if (to5[i + 1] == ',' ) {
          continue;
	}
        nospace_topic1[j] = to5[1];
	j++;
        nospace_topic1[j] = atoi("t");
	j++;
        nospace_topic1[j] = atoi("O");
	j++;
        nospace_topic1[j] = k;
	j++;
	continue;
      }
      nospace_topic1[j] = to5[i];
      j++;
    }
    strcpy(to5,nospace_topic1);

    strcat(topic_string,"_tO5");
    strcat(topic_string,to5);

  }


  GD_S(GNUT_HTTP_DEBUG, "BBBBBBBBbb   nospacetopic1=BBB");
  GD_S(GNUT_HTTP_DEBUG, nospace_topic1);
  GD_S(GNUT_HTTP_DEBUG, "BBB\n");

  GD_S(GNUT_HTTP_DEBUG, "Found timestamps  GGGgGgGg ");
  GD_I(GNUT_HTTP_DEBUG, ret2);
  GD_S(GNUT_HTTP_DEBUG, ":");
  GD_S(GNUT_HTTP_DEBUG, ts);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  strcat(create_string," ");
  strcat(create_string,ts);

  strcat(create_string," ");
  strcat(create_string,cc);

  strcat(create_string," ");
  strcat(create_string,pc);

  strcat(create_string," ");
  strcat(create_string,to1);

  strcat(create_string," ");
  strcat(create_string,to2);

  GD_S(GNUT_HTTP_DEBUG, "YY3YYYYYYYYYYYYYYYYYYY==  XxX");
  GD_S(GNUT_HTTP_DEBUG, create_string);
  GD_S(GNUT_HTTP_DEBUG, "XxX\n");

  sprintf(leafname,"LEAF_tS%s_pL%s_cC%s_pC%s%s.txt",ts,pl,cc,pc,topic_string);

  GD_S(GNUT_HTTP_DEBUG, "BBBBBBBBbb   leafname=BBB");
  GD_S(GNUT_HTTP_DEBUG, leafname);
  GD_S(GNUT_HTTP_DEBUG, "BBB\n");


  tmp = conf_get_str("share_paths");
  fpath = expand_path(tmp);
  sprintf(filename,"%s/%s",fpath,leafname);
  file = fopen(filename, "w+");
  if (!file) {
    printf("Couldn't open file \"%s\" for writing.\n", filename);
    return 0;
  }
    fprintf(file, "%s\n",cn);
    fflush(file);
    fclose(file);

    parse_command("scan"); /* rescan shared files for new LEAF */
  } else {
    create_string = 0;
  }

  GD_S(GNUT_HTTP_DEBUG, "Writing headers\n");

  http_response_header_add(req, "Content-Type", "text/html");
  http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
  http_response_header_add(req, "Connection", "close");
  http_response_header_write(req);

  GD_S(GNUT_HTTP_DEBUG, "Writing create form\n");

  gnut_http_write_top(req);
  gnut_http_write_form_create(req, 111, 222);

  gnut_http_write_bottom(req);

  return 0;
}

/* gnut_http_file_remove function to remove a leaf through the http gui */
int gnut_http_file_remove(gnut_transfer *gt, http_request_t *req)
{
/*
  Gnut_List *list;
  query_resp *a_qrp;
  */
  char *create_string = 0;
  char *query_string;

  /*
  char c_port[32];
  char c_address[BUFSZ1];
  unsigned int i_port;
  int n, i1;
  */
  int i;
  char guid[16];
  http_cookie_t *cookp;

  int ret2,ret3,ret4,ret5;
  char value[BUFSZ1] = "", ts[BUFSZ1] = "", pl[BUFSZ1] = "", cc[BUFSZ1] = "";
  char *leafname;




  GD_S(GNUT_HTTP_DEBUG, "gnut_http_file_remove IN\n");


  /* Check to see if we got a GUID in a cookie already... */
  cookp = http_cookie_new();

  /* Create a new guid and set it as a cookie to the browser. */
  GD_S(1,"creating new cookie\n");

  cookp->value = ymaloc(sizeof(char)*33, 328);

  memcpy(guid, conf_get_str("mac"), 6);
  for (i=6; i<16; i++) {
    guid[i] = rand();
  }

  gnut_guid_to_string(cookp->value, guid);
  http_response_set_cookie(req, cookp);
  GD_S(GNUT_HTTP_DEBUG, "Setting the guid in a cookie ");
  GD_S(GNUT_HTTP_DEBUG, cookp->value);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  http_cookie_delete(cookp);


  query_string = 0;

  /* Get the CGI string out of the query. */
  query_string = strstr(req->url_file, "/remove/?");
  if (query_string) {

    query_string += 9;

    leafname = strstr(query_string, "LEAF");
    strncat( value, leafname, strlen(leafname) -4 );

      char *ftmp = 0;
      char dp[512];
      char sp[512];

      ftmp = conf_get_str("share_paths");
      sprintf(sp,"%s/%s",expand_path(ftmp),value);

      ftmp = conf_get_str("download_path");
      sprintf(dp,"%s/%s",expand_path(ftmp),value);

      rename(sp,dp);

      parse_command("scan");
      printf("leaf %s now removed\n", value);


    GD_S(GNUT_HTTP_DEBUG, "value==  XxX");
    GD_S(GNUT_HTTP_DEBUG, value);
    GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");

    GD_S(GNUT_HTTP_DEBUG, "leafname==  XxX");
    GD_S(GNUT_HTTP_DEBUG, leafname);
    GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");

    ret3 = 0;

    ret2 = cgi_parse(query_string, "remove1", ts);
    if (  !strcmp(ts, "") ) {
      ret3 = cgi_parse(query_string, "remove2", ts);
    }

    strncpy(pl,"earth",5);  /* static for now */

    ret4 = cgi_parse(query_string, "remove3", cc);
    if (  !strcmp(cc, "") ) {
      ret5 = cgi_parse(query_string, "remove4", cc);
    }

      GD_S(GNUT_HTTP_DEBUG, "ret2==  XxX");
      GD_I(GNUT_HTTP_DEBUG, ret2);
      GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");
      GD_S(GNUT_HTTP_DEBUG, "ret3==  XxX");
      GD_I(GNUT_HTTP_DEBUG, ret3);
      GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");
      GD_S(GNUT_HTTP_DEBUG, "ts==  XxX");
      GD_S(GNUT_HTTP_DEBUG, ts);
      GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");
      GD_S(GNUT_HTTP_DEBUG, "cc==  XxX");
      GD_S(GNUT_HTTP_DEBUG, cc);
      GD_S(GNUT_HTTP_DEBUG, "XxX\n  ");

      // processing values
    create_string = value;

  } else {
    create_string = 0;
  }

  /*
  GD_S(GNUT_HTTP_DEBUG, "Writing headers\n");

  http_response_header_add(req, "Content-Type", "text/html");
  http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
  http_response_header_add(req, "Connection", "close");
  http_response_header_write(req);
*/

  GD_S(GNUT_HTTP_DEBUG, "Writing create form\n");

  gnut_http_file_list(gt, req);

  return 0;
}

/* gnut_http_file_options function to show the default options page */
int gnut_http_file_options(gnut_transfer *gt, http_request_t *req)
{

  int retval;

  int optionschanged = 0;

  char value[256] = "";

  char *query_string;

  char connectcom2[32];


  query_string = 0;

  /* Get the CGI string out of the query. */
  query_string = strstr(req->url_file, "/options/?");
  if (query_string) {

    query_string += 10;

    retval = cgi_parse(query_string, "gg_default_language", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_default_language", value); optionschanged = 1 ;  }

    retval = cgi_parse(query_string, "gg_default_planet", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_default_planet", value); optionschanged = 1 ;  }

    retval = cgi_parse(query_string, "gg_default_cc", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_default_cc", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_default_postcode", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_default_postcode", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top1", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top1", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top2", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top2", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top3", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top3", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top4", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top4", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top5", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top5", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_def_top6", value);
    if ( strlen(value) >= 1 ) {  conf_set_str("gg_def_top6", value);  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_threshold", value);
    if ( strlen(value) >= 1 ) {  conf_set_int("gg_threshold", atoi(value));  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "gg_dl_timeout", value);
    if ( strlen(value) >= 1 ) {  conf_set_int("gg_dl_timeout", atoi(value));  optionschanged = 1 ; }

    retval = cgi_parse(query_string, "max_results", value);
    if ( strlen(value) >= 1 ) {  conf_set_int("max_results", atoi(value));  optionschanged = 1 ; }


    retval = cgi_parse(query_string, "connect_ip", value);

    if ( strlen(value) >= 1 ) {
      sprintf(connectcom2,"%s %s","open ",value);
      parse_command(connectcom2);
    }

    if ( optionschanged > 0 ) {
      com_save(0);
    }

    gnut_http_file_search(gt, req);


  } else {


    GD_S(GNUT_HTTP_DEBUG, "gnut_http_options IN\n");

    GD_S(GNUT_HTTP_DEBUG, "Writing headers\n");
    http_response_header_add(req, "Content-Type", "text/html");
    http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
    http_response_header_add(req, "Connection", "close");
    http_response_header_write(req);

    GD_S(GNUT_HTTP_DEBUG, "Writing top\n");

    gnut_http_write_top(req);
    gnut_http_write_form_options(req, 111, 222);
    gnut_http_write_bottom(req);

  }

  return 0;

}

/* gnut_http_file_localconnect function to connect to local gumnut instances */
int gnut_http_file_localconnect(gnut_transfer *gt, http_request_t *req)
{

  int retval;

  char value[16] = "";

  char connectcom[32];

  char *query_string;


  query_string = 0;

  /* Get the CGI string out of the query. */
  query_string = strstr(req->url_file, "/localconnect/?");
  if (query_string) {

    query_string += 15;

  }


  retval = cgi_parse(query_string, "lanconnect", value);

  if ( atoi(value) == 3 ) {
    parse_command("netconnect");
  } else {
    sprintf(connectcom,"%s %s","lanconnect ",value);

    parse_command(connectcom);
  }

  gnut_http_file_search(gt, req);


  return 0;

}

int gnut_http_file_search(gnut_transfer *gt, http_request_t *req)
{
  Gnut_List *list;
  query_resp *a_qrp;
  char *search_string = 0;
  char *query_string = 0;
  int ret;
  char c_port[32];
  char c_address[BUFSZ1];
  unsigned int i_port;
  char value[BUFSZ1];
#ifdef GNUT_HTTP_SEARCH
  char c_gnut_min_returns[32];
  char c_gnut_time_to_wait[32];
  int gnut_time_to_wait;
  int gnut_min_returns;
  int n;
  int count;
#endif
  int i, i1;
  char guid[16];
  http_cookie_t *cookp;

  int ret_support = 0;
  char value_support[BUFSZ1];
  char *leafname;
  char searchname[96];

  int ret2;
  char value2[BUFSZ1];
  int ret3;
  char value3[BUFSZ1];
  int ret4;
  char value4[BUFSZ1];
  int ret5;
  char value5[BUFSZ1];
  int ret6;
  char value6[BUFSZ1];
  int ret7;
  char value7[BUFSZ1];
  int ret8;
  char value8[BUFSZ1];
  int ret9;
  char value9[BUFSZ1];
  int ret10;
  char value10[BUFSZ1];

  gnut_time_to_wait = 60;
  gnut_min_returns = 20;

  int supportorsort = 0;

  int ret_display = 0;
  char value_display[BUFSZ1];

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_file_search IN\n");

  /* Check to see if we got a GUID in a cookie already... */
  cookp = http_cookie_new();

  /* Create a new guid and set it as a cookie to the browser. */
  GD_S(1,"creating new cookie\n");

  cookp->value = ymaloc(sizeof(char)*33, 328);

  memcpy(guid, conf_get_str("mac"), 6);
  for (i=6; i<16; i++) {
    guid[i] = rand();
  }

  gnut_guid_to_string(cookp->value, guid);
  http_response_set_cookie(req, cookp);
  GD_S(GNUT_HTTP_DEBUG, "Setting the guid in a cookie ");
  GD_S(GNUT_HTTP_DEBUG, cookp->value);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  http_cookie_delete(cookp);

  /* Get the CGI string out of the query. */
  query_string = strstr(req->url_file, "/search/?");
  if (query_string) {
    query_string += 9;

    ret_display = cgi_parse(query_string, "display", value_display);

    if ( (! ret_display) ) {
     query_string = conf_get_str("gg_last_query");
     guidispindex = atoi(value_display);

    } else {
       guidispindex = 1;
    }

    ret = cgi_parse(query_string, "query", value);

    GD_S(GNUT_HTTP_DEBUG, "Found ");
    GD_I(GNUT_HTTP_DEBUG, ret);
    GD_S(GNUT_HTTP_DEBUG, ":");
    GD_S(GNUT_HTTP_DEBUG, value);
    GD_S(GNUT_HTTP_DEBUG, "\n");
    if (!ret) {
      search_string = value;

      char qs[512];
      sprintf(qs,"%s",query_string);
      conf_set_str("gg_last_query",qs);
      com_save(0);
    }



    ret_support = cgi_parse(query_string, "support", value_support);


    if (! ret_support) {
      supportorsort = 1;

      leafname = strstr(req->url_file, "/search/?support=");
      if (leafname) {
        leafname += 17;
      }

      char *ftmp = 0;
      char sp[256];

      ftmp = conf_get_str("share_paths");
      sprintf(sp,"%s/%s",expand_path(ftmp),leafname);

      if (! access(sp,R_OK) ) {  /* does the leaf exist in share path */
        printf("LEAF already mirrored at:\n%s\n",sp);
        gnut_http_file_list(gt, req);
        return 0;
      }

      printf("leafname = %s\n", leafname);

      conf_set_int("wait_after_find",0);

      char tmpcom[256];

      //search queries over 78 bytes are dumped, trim ours to 77.
      strncpy(searchname, leafname, 77  );

      sprintf(tmpcom,"find %s",searchname);
      parse_command(tmpcom);

      sleep(1);

      parse_command("displaytree");

      sleep(1);


      char dp[256];

      ftmp = conf_get_str("download_path");
      sprintf(dp,"%s/%s",expand_path(ftmp),leafname);

      rename(dp,sp);

      parse_command("scan");
      printf("LEAF now supported at:\n%s\n",sp);

      conf_set_int("wait_after_find",1);

    }

    if ( ! ret ) {
      query_string = conf_get_str("gg_last_query");
      ret = cgi_parse(query_string, "query", value);
      search_string = value;
    }

    int ret_sort = 0;
    char value_sortby[BUFSZ1];
    char *cursortby = 0;
    char *cursortord = 0;
    char *neword = 0;

    ret_sort = cgi_parse(query_string, "sort", value_sortby);

    if (! ret_sort) {
      supportorsort = 1;

      cursortby = conf_get_str("gg_sort_by");
      cursortord = conf_get_str("gg_sort_order");

      neword = "norm";

//	PS("sortable columns dev\n");
//	PS(value_sortby);
//	PS(cursortby);
//	PS(cursortord);
//	PS(neword);
//	PI(ret_sort);

      if (! strcmp( cursortby   , value_sortby ) ) {

        if (! strcmp( cursortord   , "norm" ) ) {
          neword = "rev";
        }

      } else {

        conf_set_str( "gg_sort_by", value_sortby  );

      }

      conf_set_str( "gg_sort_order", neword );

      com_save(0);


      cursortby = conf_get_str("gg_sort_by");

//	PS("sortable columns dev\n");
//	PS(value_sortby);
//	PS(cursortby);
//	PS(cursortord);
//	PS(neword);
//	PI(ret_sort);

//	PS(search_string);

        gnut_http_file_list(gt, req);

      //get last search that had a query and put into gt or req  and search here
       //     gnut_http_file_search(gt, req);

      return 0;

    }


#ifdef GNUT_HTTP_SEARCH
    ret = cgi_parse(query_string, "returns", c_gnut_min_returns);

    if (ret) {
      gnut_min_returns = 20;
    } else {
      gnut_min_returns = atoi(c_gnut_min_returns);
    }
    if (gnut_min_returns > 1200) {
      gnut_min_returns = 1200;
    } else if (gnut_min_returns < 1) {
      gnut_min_returns = 1;
    }
    GD_S(GNUT_HTTP_DEBUG, "Min Returns: ");
    GD_I(GNUT_HTTP_DEBUG, gnut_min_returns);
    GD_S(GNUT_HTTP_DEBUG, "\n");

    ret = cgi_parse(query_string, "wait", c_gnut_time_to_wait);
    if (ret) {
      gnut_time_to_wait = 3;
    } else {
      gnut_time_to_wait = atoi(c_gnut_time_to_wait);
    }
    if (gnut_time_to_wait >600) {
      gnut_time_to_wait = 600;
    } else if (gnut_time_to_wait < 3) {
      gnut_time_to_wait = 3;
    }
    GD_S(GNUT_HTTP_DEBUG, "Max time to wait for returns ");
    GD_I(GNUT_HTTP_DEBUG, gnut_time_to_wait);
    GD_S(GNUT_HTTP_DEBUG, "\n");


#endif
  } else {
    search_string = 0;
  }

  GD_S(GNUT_HTTP_DEBUG, "Writing headers\n");

  http_response_header_add(req, "Content-Type", "text/html");
  http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
  http_response_header_add(req, "Connection", "close");
  http_response_header_write(req);

  GD_S(GNUT_HTTP_DEBUG, "Writing search form\n");


  gnut_http_write_top(req);
  gnut_http_write_form(req, gnut_time_to_wait, gnut_min_returns);

  GD_S(GNUT_HTTP_DEBUG, "1111111Writing search form\n");


  if (search_string  ) {
    if ( (strlen(search_string) < 1) & ret_support ) {
      ret10 = cgi_parse(query_string, "recent_query", value10);

      if( strlen(value10) > 3 ) {
        strcat(search_string,value10);
        conf_set_str("gg_last_query",query_string);
        com_save(0);
      } else {
        strcat(search_string,"Leaf");
      }
    }
  }


  GD_S(GNUT_HTTP_DEBUG, "2222222Writing search form\n");
  /* End this now if there's no meaningful search to perform. */
  if ((!search_string) || (strlen(search_string) < 4)) {
    http_write_str(req, "<br><br><b><i>");
    if (search_string) {
      http_write_str(req, UI_HT_4CHARS_ERROR);
    }
    http_write_str(req, "</b></i>\r\n");
    gnut_http_write_bottom(req);
    return 0;
  }


  GD_S(GNUT_HTTP_DEBUG, "3333333Writing search form\n");

  ret2 = cgi_parse(query_string, "timestamps", value2);
  ret3 = cgi_parse(query_string, "time_stamps", value3);
  ret4 = cgi_parse(query_string, "countrycodes", value4);
  ret5 = cgi_parse(query_string, "recent_cc", value5);
  ret6 = cgi_parse(query_string, "postcodes", value6);
  ret7 = cgi_parse(query_string, "recent_pc", value7);
  ret8 = cgi_parse(query_string, "threshold", value8);
  ret9 = cgi_parse(query_string, "recent_threshold", value9);


  char threshcom[32];

  if  (strcmp(value8, "")) {
    sprintf(threshcom, "threshold %s", value8);
    parse_command(threshcom);
  } else if (strcmp(value9, "")) {
    sprintf(threshcom, "threshold %s", value9);
    parse_command(threshcom);
  }


  GD_S(GNUT_HTTP_DEBUG, "Found timestamps  GGGgGgGg ");
  GD_I(GNUT_HTTP_DEBUG, ret2);
  GD_S(GNUT_HTTP_DEBUG, ":");
  GD_S(GNUT_HTTP_DEBUG, value2);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  if (strcmp(value2, "")) {
    strcat(search_string," ");
    strcat(search_string,value2);  //timestamps
  } else if (strcmp(value3, "")) {
      time_t tim;
      char ts[256] = "";
      tim = time(0);
    if (!strcmp(value3, "Today")) {

      strftime(ts,16,"%Y%m%d",localtime(&tim));
      strcat(search_string," _tS");
      strcat(search_string,ts);  //time_stamps


/*
    } else if (!strcmp(value3, "Past Week")) {
      char ts_day[256] = "";
      strftime(ts,16,"%Y%m%d",localtime(&tim));
      strftime(ts_day,16,"%e",localtime(&tim));
  // time stamp drop down box   code below to check functionality
      if ( (1 <= atoi(ts_day))  & (atoi(ts_day) <= 4) ) {
        strftime(ts,16,"%Y",localtime(&tim));
      } else if ( (5 <= atoi(ts_day)) & (atoi(ts_day) <= 9) ) {
        if ( (5 <= atoi(ts_day))  & (atoi(ts_day) <= 9) ) {
          strftime(ts,16,"%Y%m",localtime(&tim));
        } else if ( (10 <= atoi(ts_day)) & (atoi(ts_day) <= 16) ) {
          strftime(ts,16,"%Y%m",localtime(&tim));
          strcat(ts,"1");
        }
      } else if ( (17 <= atoi(ts_day)) & (atoi(ts_day) <= 16) ) {
        strftime(ts,16,"%Y%m",localtime(&tim));
        strcat(ts,"1");
      }
*/

 ///     strftime(ts,16,"%Y%m%d",localtime(&tim));
 ///     strcat(search_string," ");
 ///     strcat(search_string,ts);  //time_stamps
 ///     search_string[strlen(search_string)-1] = '\0';

    } else if (!strcmp(value3, "This Month")) {
      strftime(ts,16,"%Y%m",localtime(&tim));
      strcat(search_string," _tS");
      strcat(search_string,ts);
    } else if (!strcmp(value3, "Last Month")) {
      strftime(ts,16,"%m",localtime(&tim));
      if ( atoi(ts) == 1 ) {
        strcat(search_string," _tS");
        strftime(ts,16,"%Y",localtime(&tim));
        sprintf(ts, "%d", atoi(ts) - 1);
        strcat(search_string,ts);
   //      sprintf(ts, "%d", "12");
 //        strcpy(ts,"12");
	///strcat(search_string,ts);
        strcat(search_string,"12");
      } else {
        strcat(search_string," _tS");
        strftime(ts,16,"%Y",localtime(&tim));
        strcat(search_string,ts);
        strftime(ts,16,"%m",localtime(&tim));
        sprintf(ts, "%02d", atoi(ts) - 1);
        strcat(search_string,ts);
      }
    } else {
      strcat(search_string," ");
      strcat(search_string,value3);  //time_stamps input value
    }
  }

  if (strcmp(value4, "")) {
    strcat(search_string," ");
    strcat(search_string,value4);  //countrycodes
  } else if (strcmp(value5, "")) {
    strcat(search_string," ");
    strcat(search_string,value5);  //recent_cc
  }
  if (strcmp(value6, "")) {
    strcat(search_string," ");
    strcat(search_string,value6);  //postcodes
  } else if (strcmp(value7, "")) {
    strcat(search_string," ");
    strcat(search_string,value7);  //recent_pc
  }


  GD_S(GNUT_HTTP_DEBUG, "Starting new Gnutella-net query for ");
  GD_S(GNUT_HTTP_DEBUG, search_string);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  if (supportorsort) {
    query_string = conf_get_str("gg_last_query");
  }

  gnut_http_query_new(gt, search_string, guid);

  http_write_str(req, "<br>");
  http_write_str(req, "<hr />");
  http_write_str(req, "<br>");
  http_write_str(req, "<font size=+3>");
  http_write_va(req, BUFSZ1, UI_HT_SEARCH_RESULTS, search_string);
  http_write_str(req, "</font>");

  int thresh = conf_get_int("gg_threshold");
  http_write_va(req, BUFSZ1, " &nbsp &nbsp &nbsp   threshold = %i ", thresh);

// unnecessary //  http_write_va(req, BUFSZ1, UI_HT_PLEASE_WAIT, gnut_time_to_wait);
  http_write_str(req, "<br><br>\r\n");

  /* Wait for  time_to_wait seconds,
   * or min_returns requests, whichever
   * is first. - By default, this is not a good
   * thing to do, as it's a little cumbersome.  */
#ifdef GNUT_HTTP_SEARCH
  count = 0;


  while (count < gnut_time_to_wait) {
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "594 calling gnut_http_query_get IN\n");
    list = gnut_http_query_get(gt, guid);
    if (list) {
      n = gnut_list_size(list);
    } else {
      /* No list means no results have come back yet */
      n = 0;
    }
    if (n >= gnut_min_returns) {
      count = gnut_time_to_wait;
    }
    sleep(1); count++;

    /* This keeps some browsers from printing "stalled" in the status bar */
    http_write_str(req, " ");
  }
#endif

  int cliport;

  cliport = conf_get_int("local_port");

  http_write_str(req, "<br>");

  http_write_str(req, "<table><tr bgcolor=\"cbcbce\"><td>");

  http_write_str(req, "<font size=+1> # </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Filesize </font>");
 /// disab ///   http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?sort=filesize\">&nbsp filesize &nbsp</a>", "127.0.0.1", cliport);
  http_write_str(req, "</td><td>");
//  http_write_str(req, "..speed..");
//  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Support-Leaf </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Local </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Weighting </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Date Stamp </font>");
  http_write_str(req, "</td><td>");
//  http_write_str(req, "<font size=+1> Planet </font>");
//  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Country </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Postcode </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Topic 1 </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Topic 2 </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Topic 3 </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Topic 4 </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> Topic 5 </font>");

  http_write_str(req, "</td></tr>");


  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "calling gnut_http_query_get IN\n");
  list = gnut_http_query_get(gt, guid);
      n = gnut_list_size(list);

  Gnut_List *newlist = NULL;

  while(1) {
    if (!list) {
      break;
    }
    a_qrp = list->data;
    if (!a_qrp) {
      break;
    }

   char* curleafname;

   curleafname = a_qrp->qr_ndata;

   Gnut_List *gltmp;
   query_resp *b_qrp;
   int leafcount = 0;

   for (gltmp=list; gltmp; gltmp=gltmp->next) {
     b_qrp = gltmp->data;

       // if leafs match, increment leafcount
     if ( !strcmp( b_qrp->qr_ndata, curleafname ) ) {
       leafcount++;
     }
   }

   a_qrp->qr_ggnumhits = leafcount;

   int needtoadd = 1;

   Gnut_List *gltmp2;
   query_resp *c_qrp;

   for (gltmp2=newlist; gltmp2; gltmp2=gltmp2->next) {
     c_qrp = gltmp2->data;

     if ( !strcmp( c_qrp->qr_ndata, curleafname ) ) {
       needtoadd = 0;
     }

   }

  // int thresh = conf_get_int("gg_threshold");

   if ( needtoadd == 1  && leafcount >= thresh ) {
     newlist = gnut_list_append(newlist,a_qrp,42);
   }

   list = gnut_list_next(list);

  }

  list = newlist;

  i1 = 0;


  while(1) {

    if (!list) {
      break;
    }

    a_qrp = list->data;
    if (!a_qrp) {
      break;
    }


    i_port = GET_LEU16(a_qrp->qr_port_le, 22);

    sprintf(c_port, "%i", i_port);
    sprintf(c_address, "%i.%i.%i.%i", a_qrp->qr_ip[0], a_qrp->qr_ip[1],
	    a_qrp->qr_ip[2], a_qrp->qr_ip[3]);

      char * name2;
      char name3[512];
      int cliport;

      cliport = conf_get_int("local_port");

      name2 = ystdup(a_qrp->qr_ndata, 168);
      make_htmlsafe(name2);

    i1++;

    if ( i1 % 2 ) {
      http_write_str(req, "<tr bgcolor=\"f3f3f3\">");
    } else {
      http_write_str(req, "<tr bgcolor=\"e1e1e1\">");
    }

      http_write_va(req, BUFSZ1, "</td><td><a href=\"http://%s:%s/get/%u/%s\">%i</a></td>", c_address, c_port, a_qrp->qr_ref, name2, i1);


    http_write_va(req, BUFSZ1, "<td>%u</td>", a_qrp->qr_size);

// taken speed out  0.1.9 -  http_write_va(req, BUFSZ1, "<td>%u</td>", a_qrp->qr_speed);

    /* Write the filename (which is also an anchor) */

    { /* 0.4.28.c15 */
      http_write_va(req, BUFSZ1, "<td><a href=\"http://%s:%d/search/?support=%s\">support</a></td><td>", "127.0.0.1", cliport, name2);

      char leafloc[1024];
      char *tmppath = 0;

      /* does LEAF exist in download path */
      tmppath = conf_get_str("share_paths");
      sprintf(leafloc,"%s/%s",expand_path(tmppath),name2);

      if ( ! access(leafloc,R_OK) ) {  /* if leaf is local in share path */
        http_write_va(req, BUFSZ1, "L", a_qrp->qr_size);
      } else {
        http_write_va(req, BUFSZ1, "R", a_qrp->qr_size);
      }


      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "%d", a_qrp->qr_ggnumhits);

      http_write_str(req, "</td><td>");




      char ts[256] = "", pl[256] = "", cc[256] = "", pc[256] = "", to1[256] = "", to2[256] = "", to3[256] = "", to4[256] = "", to5[256] = "", alltopics[256] = "", topbuf[256] = "" ;
      char* name4;
      int k,tn;


      name4 = strstr(name2,"_tO1");
      name4 += 4;

      strcpy(name3,name2);

      sscanf(name3,"LEAF_tS%[0-9]_pL%[^_]_cC%[^_]_pC%[^_]_tO1%s",ts,pl,cc,pc,alltopics);

      tn = 1;

      for (k=0;k <= strlen(name4);k++) {

        if ( (k == strlen(name4) - 4) || (  name4[k] == '_' && name4[k + 1] == 't' && name4[k + 2] == 'O' ) ) {


          switch(tn) {
          case 1:
            topbuf[k] = atoi("\0");
            strcpy(to1,topbuf);
            tn++;
            break;
          case 2:
            topbuf[k] = atoi("\0");
            strcpy(to2,topbuf);
            tn++;
            break;
          case 3:
            topbuf[k] = atoi("\0");
            strcpy(to3,topbuf);
            tn++;
            break;
          case 4:
            topbuf[k] = atoi("\0");
            strcpy(to4,topbuf);
            tn++;
            break;
          case 5:
            topbuf[k] = atoi("\0");
            strcpy(to5,topbuf);
            tn++;
            break;
          }

          name4 += (k + 4);
          k = -1;

        } else {
          topbuf[k] = name4[k];
        }

      }


      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, ts);

      http_write_str(req, "</td><td>");

//      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, pl);
//
//      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, cc);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, pc);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, to1);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, to2);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, to3);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, to4);

      http_write_str(req, "</td><td>");

      http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%s\">%s   </a></td>", "localhost", cliport, name2, to5);


    }
    http_write_str(req, "</a>");
    http_write_str(req, "</td>");
    http_write_str(req, "</tr>");

    if ( (! ret_display)  ) {


      leafname = strstr(req->url_file, "/search/?display=");
      leafname += 17;

      if ( !strcmp( leafname, name2 ) ) {

        char dlcmd[128] = "";
        char buf[128] = "";
        FILE *fp;
        char dp[256];
        char *ff = 0;
        char *ftmp = 0;
        char *fpath = 0;
        int waittime = 0;

        char strtrans[2048] = "";

        ff = a_qrp->qr_ndata;

        ftmp = conf_get_str("download_path");
        fpath = expand_path(ftmp);
        sprintf(dp,"%s/%s",fpath,ff);

        if ( access(dp,R_OK) ) {  // leaf does not exist in download path
          strncat(buf,ff,77 );
          conf_set_str("wait_after_find","0");
          sprintf(dlcmd,"find %s",buf);
          parse_command(dlcmd);
          sleep(1);
PS(buf);
PS(ff);
PS(dlcmd);
PI(ret_display);
          parse_command("displaytree");
          conf_set_str("wait_after_find","1");
          while ( access(dp,R_OK) ) {
PI(access(dp,R_OK));
            if ( waittime <= 10 ) {
PI(waittime);
              sleep(1);   // wait until downloaded before moving leaf.
              waittime++;
            } else {
              printf("timed out waiting for leaf to download\n");
              return -1;
            }
          }
        }



        if ( i1 % 2 ) {
          http_write_str(req, "<tr bgcolor=\"f3f3f3\">");
        } else {
          http_write_str(req, "<tr bgcolor=\"e1e1e1\">");
        }

      //  http_write_str(req, "<tr bgcolor=\"e1e1e1\"><td></td><td></td><td></td><td></td><td></td><td>");
        http_write_str(req, "<td></td><td></td><td></td><td></td><td></td><td>");
     //   http_write_str(req, "<tr><td></td><td></td><td></td><td></td><td></td><td>");

        fp = fopen(dp, "r");

        while (fgets(buf, sizeof(buf), fp)) {
          http_write_va(req, BUFSZ1, "%s", buf);

          strcat(strtrans,buf);
          strcat(strtrans," ");
        }
        http_write_str(req, "</td>");
        fclose(fp);

        if ( i1 % 2 ) {
          http_write_str(req, "<tr bgcolor=\"f3f3f3\">");
        } else {
          http_write_str(req, "<tr bgcolor=\"e1e1e1\">");
        }

        http_write_str(req, "<td></td><td></td><td></td><td></td><td></td><td>");
     //   http_write_str(req, "<tr><td></td><td></td><td></td><td></td><td></td><td>");



        ftmp = conf_get_str("gg_default_language");
        http_write_va(req, BUFSZ1, "<a href=\"http://translate.google.com/translate_t?hl=%s#auto|%s|%s\">translate contents</a></td>",ftmp,ftmp, strtrans);


/*  ** forward and backward  not functioning - not displayed **
        http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%i\">forward</a>", "127.0.0.1", cliport, guidispindex + 1);
        http_write_str(req, " ");

        if (guidispindex == 1) {
          http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%i\">backward</a>", "127.0.0.1", cliport, guidispindex );
        } else {
          http_write_va(req, BUFSZ1, "<a href=\"http://%s:%d/search/?display=%i\">backward</a>", "127.0.0.1", cliport, guidispindex - 1 );
        }
*/

        http_write_str(req, "</td>");

        http_write_str(req, "<tr><td>&nbsp;</td></tr>"); 
      }
    }


    fre_str(&name2, 170);

    list = gnut_list_next(list);

    if (!list) {
      break;
    }
  }

  http_write_str(req, "</table>");

  http_write_str(req, "<br>");

  gnut_http_write_bottom(req);

  return 0;

}
#endif


#ifdef GNUT_HTTP_FILE_LIST
int gnut_http_file_list(gnut_transfer *gt, http_request_t *req)
{
  Gnut_List *list;
  share_item *si;
  int i1;

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_file_list IN\n");

  GD_S(GNUT_HTTP_DEBUG, "Writing headers\n");
  http_response_header_add(req, "Content-Type", "text/html");
  http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
  http_response_header_add(req, "Connection", "close");
  http_response_header_write(req);

  GD_S(GNUT_HTTP_DEBUG, "Writing top\n");
  gnut_http_write_top(req);


  http_write_str(req, "<font size=+3>");
  /// renamed http_write_str(req, UI_HT_FILES_ON);
  http_write_str(req, "Local files");
  http_write_str(req, "</font>");

  http_write_str(req, "<br><br>");

  http_write_str(req, "<FORM ACTION=\"/remove/\">");
  http_write_str(req, "<input value=\"");
  http_write_str(req, "Remove checked leaf");
  http_write_str(req, "\" type=\"submit\"><br>");

  http_write_str(req, "<br>");

  http_write_str(req, "<table><tr bgcolor=\"cbcbce\"><td>");
  http_write_str(req, "#");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> &nbsp Remove-Leaf &nbsp </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> &nbsp Filesize &nbsp </font>");
  http_write_str(req, "</td><td>");
  http_write_str(req, "<font size=+1> &nbsp File-Location &nbsp </font>");
  http_write_str(req, "</td></tr>");

  GD_S(GNUT_HTTP_DEBUG, "Getting share root\n");
  list = share_get_root();

  GD_S(GNUT_HTTP_DEBUG, "Writing local share list\n");
  i1 = 0;
  while(1) {
    if (!list) break;
    si = list->data;
    if (!si) break;

    i1++;

    if ( i1 % 2 ) {
      http_write_str(req, "<tr bgcolor=\"f3f3f3\">");
    } else {
      http_write_str(req, "<tr bgcolor=\"e1e1e1\">");
    }


    http_write_va(req, BUFSZ1, "<td>%i</td>", i1);


    { /* 0.4.28.c15 */
      char * name2;



      name2 = ystdup(si->path, 238);

      make_htmlsafe(name2);
      http_write_va(req, BUFSZ1, "<td><input type=\"checkbox\" NAME=\"remove%i=%s_\" SIZE=10>&nbsp;</td>", i1, name2);

      http_write_va(req, BUFSZ1, "<td>%u</td>", si->size);
      http_write_str(req, "<td>");
      http_write_va(req, BUFSZ1, "<a href=\"/get/%u/%s\">", si->ref, name2);
      http_write_str(req, name2);
      fre_str(&name2, 239);
    }
    http_write_str(req, "</a>");

    http_write_str(req, "</td>");

    http_write_str(req, "</tr>\n");

    list = gnut_list_next(list);
  }
  http_write_str(req, "</table>");

  http_write_str(req, "<br>");

  http_write_str(req, "<input value=\"");
  http_write_str(req, "Remove checked leaf");
  http_write_str(req, "\" type=\"submit\"><br>");

  http_write_str(req, "</FORM>");

  GD_S(GNUT_HTTP_DEBUG, "Writing bottom of HTML page.\n");
  gnut_http_write_bottom(req);

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_file_list OUT\n");

  return 0;
}
#endif

int parse_http_request(char *request, int32 *ref, char **name)
{
  uint32 iref;
  char *ptr, *ptr2;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse_http.request IN - ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, request);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

  ptr = strstr(request,"/get/");
  if (ptr == 0) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no /get/ in url\n");
    return -2;
  }

  /* Danger will robinson. Unsafe pointer manipulations. */
  ptr += 5;
  ptr2 = strchr(ptr,'/');
  if (ptr2 == 0) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no reference\n");
    return -3;
  }

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "located end slash ptr = ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, ptr);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "located end slash ptr2 = ");
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, ptr2);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

  /* Get file refnum */
  iref = gl_stou32(ptr, 0);

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "located ref num: ");
  GD_U(GNUT_HTTP_INTERNALS_DEBUG, iref);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

  /* By the Gnutellanet protocol standard, the name is irrelevant -- but
   * our caller might want to use it for logging purposes. */
  *name = ystdup(ptr2, 464);
  *ref = iref;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse_http.request OUT\n");
  return 0;
}


/*
 *
 * ***  END modified http mini server with gumnut search capability  ***
 *
 */


static int parse_range(char *buf, uint32 size, uint32 *rangemin,
		       uint32 *rangemax)
{
  int i;
  uint32 num;
  uint32 num2;
  char *ptr;

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range IN\n");

  for (i=6; i<strlen(buf) && !G_ISDIGIT(buf[i]) && buf[i]!='-'; i++)
    ;

  if (i==strlen(buf)) {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range OUT - full length, aparently\n");

    return 0;
  }

  if (buf[i]=='-') {
    num = gl_stou32(&buf[i+1], 0);
    if (num <= size) {
      *rangemin = size - num;
    } else {
      *rangemin = 0;
    }
    *rangemax = size - 1;
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range OUT - 1 - min=");
    GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemin);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, ", max=");
    GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemax);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

    return 1;
  }

  num = gl_stou32(&buf[i], &ptr);

  if ((*ptr) != '-') {
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range OUT - full length, aparently - bad range\n");

    return 0;
  }

  ptr++;

  if (!G_ISDIGIT(*ptr)) {
    *rangemin = num;
    *rangemax = size - 1;
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range OUT - 2 - min=");
    GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemin);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, ", max=");
    GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemax);
    GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

    return 1;
  }

  num2 = gl_stou32(ptr, 0);

  *rangemin = num;
  if (num2 < size) {
    *rangemax = num2;
  } else {
    *rangemax = size - 1;
  }

  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "parse.range OUT - 3 - min=");
  GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemin);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, ", max=");
  GD_U(GNUT_HTTP_INTERNALS_DEBUG, *rangemax);
  GD_S(GNUT_HTTP_INTERNALS_DEBUG, "\n");

  return 1;
}

/* This routine calculates the value we should send for our upload
   speed in query responses, based on all the relevant settings and
   the upload rate history. */
void update_urate()
{
  int i;
  uint32 best;
  int32 uc;

  /* The last-resort default is half the "speed" setting, on the premise
     that any upload would at least be competing with the GnutellaNet
     connections  */
  best_urate = gc_speed * 64; /* 0128/2, the 0128 is to convert kbits/sec to
                                 bytes/sec */
  urate_measured = 0; /* The speed given by the user is considered
			 "unverified" */
  urate_clipped = 0;

  /* If there is an upload history, take the best rate from that history */
  best = 0;
  for(i=0; i<10; i++) {
    if (urate_history[i] > best) {
      best = urate_history[i];
    }
  }
  if (best) {
    best_urate = best;
    urate_measured = 1;
  }

  /* Advertised measured speed cannot exceed either of the bandwidth limits */
  uc = conf_get_int("default_upload_cap");
  if ((uc > 0) && (uc < best_urate)) {
    best_urate = uc;
    urate_measured = 1;
    urate_clipped = 1;
  }

  uc = conf_get_int("global_bandwidth_cap");
  if ((uc > 0) && (uc < best_urate)) {
    best_urate = uc;
    urate_measured = 1;
    urate_clipped = 1;
  }
}

int gnut_http_serve_file(gnut_transfer *gt, http_request_t *req)
{
  uint32 ref;
  //int32 ref;          unsigned int possible fix    
  int ret;
  share_item *si = 0;
  char *buf = 0;
  uint32 partial = 0;
  uint32 rangemin = 0;
  uint32 rangemax = 0;
  char *range;

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_serve_file IN\n");
  range = http_request_find_header(req, "Range");

  /* Parse to find the actual file we're looking for. */
  ret = parse_http_request(req->url_file, &ref, &gt->fname);
  GD_S(GNUT_HTTP_DEBUG, "parse_http returned ");
  GD_I(GNUT_HTTP_DEBUG, ret);
  GD_S(GNUT_HTTP_DEBUG, " ref=");
  GD_U(GNUT_HTTP_DEBUG, ref);
  GD_S(GNUT_HTTP_DEBUG, " name=");
  GD_S(GNUT_HTTP_DEBUG, gt->fname);
  GD_S(GNUT_HTTP_DEBUG, "\n");

  if (ref) {
    si = share_find(ref);
  }

  /* this next bit is not very intuitive, but I didn't want to have
   * to have 3 separate instances of cleanup code... */
  if (ret<0 || si==0) {
    GD_S(GNUT_HTTP_DEBUG,"error with something!\n");
    if (si) {
      fre_str(&(si->path), 171);
      fre_str(&(si->lc_path), 172);
      fre_str(&(si->fpath), 173);
      fre_si(&si, 174);
    }
    return -1;
  }

  buf = ymaloc(4096, 329);

  /* If this was a partial content request,
   * then parse the amount of the partial content. */
  if (range) {
    partial = parse_range(range, si->size, &rangemin, &rangemax);
    GD_S(3, "range min: "); GD_U(3, rangemin); GD_S(3, " max: ");
    GD_U(3, rangemax); GD_S(3, "\n");

    /* workaround Gnutella asking for a rangemin > file size  */
    if (partial && (rangemin > rangemax)) {
      partial = 0;
    }

    /* Workaround BearShare specifying a range even when they want the
       whole file */
    if (rangemin == 0) {
      partial = 0;
    }
  }

  gt->fsock = open(si->fpath, O_RDONLY);

  if (gt->fsock<0) {
    GD_S(GNUT_HTTP_DEBUG,"couldn't open something!\n");

    if (gt->fsock >= 0) {
      close_s(gt->fsock);
    }
    fre_str(&buf, 175);
    if (si) {
      fre_str(&(si->path), 176);
      fre_str(&(si->lc_path), 177);
      fre_str(&(si->fpath), 178);
      fre_si(&si, 179);
    }

    return -1;
  }

  GD_S(GNUT_HTTP_DEBUG,"writing out http reply\n");

  if (partial == 0) {
    gt->gt_total = si->size;

    sprintf(buf, "Gnut/%s", VERSION);
    http_response_header_add(req, "Server", buf);

    http_response_header_add(req, "Content-Type", "application/octet-stream");

    sprintf(buf, "%u", si->size);
    http_response_header_add(req, "Content-Length", buf);

    http_response_header_write(req);
  } else {
    gt->gt_bytes = rangemin;
    gt->gt_total = rangemax + 1;

    /* I used to have the correct header here
       HTTP 206 Partial Content                            tagok
       but the real gnutella thought this was an error so I changed it */

    http_response_header_add(req, "Content-Type", "application/octet-stream");

    sprintf(buf, "bytes %u-%u", rangemin, rangemax + 1);
    http_response_header_add(req, "Content-Range", buf);

    sprintf(buf, "%u", rangemax - rangemin + 1);
    http_response_header_add(req, "Content-Length", buf);

    http_response_header_write(req);

    lseek(gt->fsock, rangemin, SEEK_SET);
  }

  gt->gt_state = STATE_CONNECTED;

  gnut_xfer_loop(gt, 0, 0, 0);

  /* Log the completed upload */
  if (gc_transferlog) {
    time_t tim;
    struct tm *ltim;
    tim = time(0);
    ltim = localtime(&tim);

    fprintf(gc_transferlog, "%d%02d%02d.%02d%02d%02d"
	    "\ts\t%i.%i.%i.%i:%i\t%u\t%u\t%u\t%s\n",
	    1900+ltim->tm_year, 1+ltim->tm_mon, ltim->tm_mday,
	    ltim->tm_hour, ltim->tm_min, ltim->tm_sec,
	    gt->ip[0], gt->ip[1], gt->ip[2], gt->ip[3],
	    gt->gt_port, gt->gt_bytes, gt->gt_total, (gt->rate_bytes) >> 4L,
	    gt->fname);
  }

  /* Since we have accomplished an upload or partial upload, that proves
     we can upload (it doesn't matter if the upload was cancelled by the
     remote user or stopped by our user, we're just interested in
     whether uploads are possible. */
  gh_did_upload = 1;

  /* Update our upload rate history */
  if (gt->gt_bytes == gt->gt_total) {
    int i;

    /* Shift entries 0-8 to positions 1-9 */
    for(i=8; i>=0; i--) {
      urate_history[i+1] = urate_history[i];
    }
    urate_history[0] = (gt->rate_bytes) >> 4L;
    update_urate();
  }

  GD_S(GNUT_HTTP_DEBUG,"closing up transfer\n");

  fre_str(&buf, 180);

  GD_S(GNUT_HTTP_DEBUG, "gnut_http_serve_file OUT\n");
  return 0;
}

#if GNUT_HTTP_FILE_LIST || GNUT_HTTP_SEARCH
void gnut_http_write_top(http_request_t *req)
{
  uint h_num;
  float64 bytes, files;

  if (html_template_head) {
    http_write_str(req, html_template_head);
  } else {
    http_write_str(req, "<HTML>\r\n");
    http_write_str(req, "<HEAD><TITLE>");
    http_write_str(req, UI_HT_MAIN_TITLE);
    http_write_str(req, "</TITLE></HEAD>\r\n");
    http_write_str(req, "<BODY BGCOLOR=\"#FFFFFF\">");

    http_write_str(req, "<table width=\"100%\"><tr><td bgcolor=\"#99CCFF\" align=center><font size=+3>");
    http_write_str(req, UI_HT_MAIN_TITLE);
    http_write_str(req, "</font></td></tr></table>");

    http_write_str(req, UI_HT_CLIENT_WEB);
    http_write_str(req, " ");
    http_write_str(req, UI_HT_POWERED);
    http_write_str(req, VERSION);

    http_write_str(req, "\r\n<br>\r\n");

    http_write_str(req, "<a href=\"/\">");
    http_write_str(req, "Get Files On This Server");
    http_write_str(req, "</a> | \r\n");
    http_write_str(req, "<a href=\"/search/\">");
    http_write_str(req, UI_HT_SEARCH_THE);
    http_write_str(req, "</a>");

    http_write_str(req, "\r\n<br>\r\n");

    http_write_str(req, UI_HT_RELATED);

    http_write_str(req, "<a href=\"http://");
    http_write_str(req, UI_HT_GNUTELLA_URL);
    http_write_str(req, "/\">");
    http_write_str(req, UI_HT_GNUTELLA_URL);
    http_write_str(req, "</a> ");

    http_write_str(req, "<a href=\"http://");
    http_write_str(req, UI_HT_GNUT_URL);
    http_write_str(req, "/\">");
    http_write_str(req, UI_HT_GNUT_URL);
    http_write_str(req, "</a>");

    http_write_str(req, "\r\n<br>\r\n");

    http_write_str(req, UI_HT_NET_STATS);
    host_totals(&h_num, &files, &bytes);
    http_write_va(req, BUFSZ1, UI_HT_STAT_HOSTS, h_num);
    http_write_va(req, BUFSZ1, UI_HT_STAT_FILES, files);
    http_write_va(req, BUFSZ1, UI_HT_STAT_BYTES, bytes / 1000000.0);

    http_write_str(req, "\r\n<br>\r\n");
  }

    int cliport;
    cliport = conf_get_int("local_port");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/\">List Local Files</a> &nbsp  | &nbsp ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/create/\">Create A New Proposal</a> &nbsp | &nbsp ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/search/\">Search The Network</a> &nbsp | &nbsp ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/options/\">Options</a> &nbsp | &nbsp ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/localconnect/?lanconnect=3\">Connect To Online Network</a>");

    http_write_str(req, "<BR>");
    http_write_str(req, "<BR>");
    http_write_str(req, "<hr>");
    http_write_str(req, "<BR>");
    http_write_str(req, "<BR>");


}

void gnut_http_write_bottom(http_request_t *req)
{

      int cliport;
      cliport = conf_get_int("local_port");

    http_write_str(req, "<BR>");
    http_write_str(req, "<BR>");
    http_write_str(req, "<hr>");
    http_write_str(req, "<BR>");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/\">List Local Files</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/create/\">Create A New Proposal</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/search/\">Search The Network</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/options/\">Options</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/localconnect/?lanconnect=0\">Connect To 192.168.0.*</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/localconnect/?lanconnect=1\">Connect To 192.168.1.*</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/localconnect/?lanconnect=2\">Connect To 192.168.2.*</a>  | ");

    http_write_str(req, "<a href=\"http://127.0.0.1:");
    http_write_va(req, BUFSZ1,"%d", cliport);
    http_write_str(req, "/localconnect/?lanconnect=3\">Connect To 10.1.1.*</a>");

    http_write_str(req, "<BR>");
    http_write_str(req, "<BR>");
    http_write_str(req, "<BR>");


  if (html_template_foot) {
    http_write_str(req,html_template_foot);
  } else {
    http_write_str(req, "</BODY>");
    http_write_str(req, "</HTML>\r\n");
  }
}

int gnut_http_write_form(http_request_t *req, int ttw, int rets)
{
  // int val, checked;

  int retf;
  char valuef[BUFSZ1];


  GD_S(2,"gnut_http_write_form entering\n");
  http_write_str(req, "<FORM ACTION=\"/search/\">");


  http_write_str(req, "<font size=+3>");
  http_write_str(req, "Search the gumnut network");
  http_write_str(req, "</font><br><br>");


  /// timestamps  ///
  http_write_str(req, "Time stamp:");
  http_write_str(req, "<input type=\"text\" NAME=\"timestamps\" SIZE=10>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"time_stamps\">");
    ///http_write_va(req, BUFSZ1, "<option VALUE=\"%i\"", val);
    http_write_str(req, "<option VALUE=\"\"");
    http_write_str(req, " selected");
    http_write_str(req, ">");
    http_write_str(req, "");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_tS2014\"");
    http_write_str(req, ">");
    http_write_str(req, "2014");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_tS2013\"");
    http_write_str(req, ">");
    http_write_str(req, "2013");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_tS2012\"");
    http_write_str(req, ">");
    http_write_str(req, "2012");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_tS2011\"");
    http_write_str(req, ">");
    http_write_str(req, "2011");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"Today\"");
    http_write_str(req, ">");
    http_write_str(req, "Today");
    http_write_str(req, "</option>");
/*    http_write_str(req, "<option VALUE=\"Past Week\"");
    http_write_str(req, ">");
    http_write_str(req, "Past Week");
    http_write_str(req, "</option>");
*/
    http_write_str(req, "<option VALUE=\"This Month\"");
    http_write_str(req, ">");
    http_write_str(req, "This Month");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"Last Month\"");
    http_write_str(req, ">");
    http_write_str(req, "Last Month");
    http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");

  /// country codes  ///
  http_write_str(req, "&nbsp &nbsp &nbsp &nbsp Country code:");
  http_write_str(req, "<input type=\"text\" NAME=\"countrycodes\" SIZE=3>&nbsp;");



  http_write_str(req, "<SELECT NAME=\"recent_cc\">");
    ///http_write_va(req, BUFSZ1, "<option VALUE=\"%i\"", val);
    http_write_str(req, "<option VALUE=\"\"");
    http_write_str(req, " selected");
    http_write_str(req, ">");
    http_write_str(req, "");
    http_write_str(req, "</option>");
    http_write_va(req, BUFSZ1, "<option VALUE=\"%s\"", conf_get_str("gg_default_cc"));
    http_write_str(req, ">");
    http_write_va(req, BUFSZ1, "%s", conf_get_str("gg_default_countrycode"));
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCau\"");
    http_write_str(req, ">");
    http_write_str(req, "au");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCbr\"");
    http_write_str(req, ">");
    http_write_str(req, "br");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCbg\"");
    http_write_str(req, ">");
    http_write_str(req, "bg");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCca\"");
    http_write_str(req, ">");
    http_write_str(req, "ca");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCcn\"");
    http_write_str(req, ">");
    http_write_str(req, "cn");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCde\"");
    http_write_str(req, ">");
    http_write_str(req, "de");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCin\"");
    http_write_str(req, ">");
    http_write_str(req, "in");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCid\"");
    http_write_str(req, ">");
    http_write_str(req, "id");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCie\"");
    http_write_str(req, ">");
    http_write_str(req, "ie");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCit\"");
    http_write_str(req, ">");
    http_write_str(req, "it");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCjp\"");
    http_write_str(req, ">");
    http_write_str(req, "jp");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCmy\"");
    http_write_str(req, ">");
    http_write_str(req, "my");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCmx\"");
    http_write_str(req, ">");
    http_write_str(req, "mx");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCnl\"");
    http_write_str(req, ">");
    http_write_str(req, "nl");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCnz\"");
    http_write_str(req, ">");
    http_write_str(req, "nz");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCno\"");
    http_write_str(req, ">");
    http_write_str(req, "no");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCru\"");
    http_write_str(req, ">");
    http_write_str(req, "ru");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCsg\"");
    http_write_str(req, ">");
    http_write_str(req, "sg");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCza\"");
    http_write_str(req, ">");
    http_write_str(req, "za");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCes\"");
    http_write_str(req, ">");
    http_write_str(req, "es");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCtt\"");
    http_write_str(req, ">");
    http_write_str(req, "tt");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCuk\"");
    http_write_str(req, ">");
    http_write_str(req, "uk");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCus\"");
    http_write_str(req, ">");
    http_write_str(req, "us");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"_cCve\"");
    http_write_str(req, ">");
    http_write_str(req, "ve");
    http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");

  http_write_str(req, "&nbsp &nbsp &nbsp &nbsp Postcode:");
  http_write_str(req, "<input type=\"text\" NAME=\"postcodes\" SIZE=5>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"recent_pc\">");
    http_write_str(req, "<option VALUE=\"\"");
    http_write_str(req, " selected");
    http_write_str(req, ">");
    http_write_str(req, "");
    http_write_str(req, "</option>");
    http_write_va(req, BUFSZ1, "<option VALUE=\"_pC%s\"", conf_get_str("gg_default_postcode"));
    http_write_str(req, ">");
    http_write_va(req, BUFSZ1, "%s", conf_get_str("gg_default_postcode"));
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"6000\"");
    http_write_str(req, ">");
    http_write_str(req, "6000");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"6000\"");
    http_write_str(req, ">");
    http_write_str(req, "6000");
    http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");
    http_write_str(req, "<br>");
    http_write_str(req, "<br>");

  http_write_str(req, "Search query:");
  http_write_str(req, "<input type=\"text\" NAME=\"query\" SIZE=30>&nbsp;");
  http_write_str(req, "<SELECT NAME=\"recent_query\">");
    http_write_str(req, "<option VALUE=\"\"");
    http_write_str(req, " selected");
    http_write_str(req, ">");
    http_write_str(req, "");
    http_write_str(req, "</option>");


  retf = cgi_parse( conf_get_str("gg_last_query"),"query", valuef);
    http_write_str(req, "<option VALUE=\"");
    http_write_str(req, valuef);
    http_write_str(req, "\">");
    http_write_str(req, valuef);
    http_write_str(req, "</option>");
  http_write_str(req, "</SELECT>");

  /// display threshold  ///
  http_write_str(req, "&nbsp &nbsp &nbsp &nbsp &nbsp Display threshold:");
  http_write_str(req, "<input type=\"text\" NAME=\"threshold\" SIZE=5>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"recent_threshold\">");
    ///http_write_va(req, BUFSZ1, "<option VALUE=\"%i\"", val);
    http_write_str(req, "<option VALUE=\"\"");
    http_write_str(req, " selected");
    http_write_str(req, ">");
    http_write_str(req, "");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"1\"");
    http_write_str(req, ">");
    http_write_str(req, "1");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"5\"");
    http_write_str(req, ">");
    http_write_str(req, "5");
    http_write_str(req, "</option>");
    http_write_str(req, "<option VALUE=\"20\"");
    http_write_str(req, ">");
    http_write_str(req, "20");
    http_write_str(req, "</option>");
  http_write_str(req, "</SELECT>");


  http_write_str(req, "<script type=\"text/javascript\"> window.onload = function() { document.getElementById('SearchButton').focus(); } </script> ");

  http_write_str(req, "&nbsp &nbsp <input id=\"SearchButton\" value=\"");
  http_write_str(req, "Search");
  http_write_str(req, "\" type=\"submit\">");

  http_write_str(req, "<br>");

  /*
#ifdef GNUT_HTTP_SEARCH
  http_write_str(req, UI_HT_WAIT_UP_TO);

  http_write_str(req, "<SELECT NAME=\"wait\">");
  checked = 0; val = 3;
  //////////ggggggg  checked = 0; val = 15;
  while (val <= 240) {
    http_write_va(req, BUFSZ1, "<option VALUE=\"%i\"", val);
    if ((checked == 0) && (val == 3)) {
   /////ggggggg  if ((checked == 0) && (val >= ttw))
      http_write_str(req, " selected");
    }
    http_write_str(req, ">");
    http_write_va(req, BUFSZ1, UI_HT_SECONDS, val);
    http_write_str(req, "</option>");
    val *= 2;
  }
  http_write_str(req, "</SELECT>");

  http_write_str(req, UI_HT_OR_UNTIL);

  http_write_str(req, "<SELECT NAME=\"returns\">");
  checked = 0; val = 5;
  while (val <= 320) { // tagok
    http_write_va(req, BUFSZ1, "<option VALUE=\"%i\"", val);
    if ((checked == 0) && (val >= rets)) {
      http_write_str(req, " selected");
    }
    http_write_str(req, ">");
    http_write_va(req, BUFSZ1, "%i", val);
    http_write_str(req, "</option>");
    val *= 2;
  }
  http_write_str(req, "</SELECT>");

  http_write_str(req, UI_HT_REPLIES_ARRIVE);
#endif
*/
  http_write_str(req, "</FORM>");

  return 0;
}

int gnut_http_write_form_options(http_request_t *req, int ttw, int rets)
{

  GD_S(2,"gnut_http_write_form_options entering\n");
  http_write_str(req, "<FORM ACTION=\"/options/\">");


  http_write_str(req, "<font size=+3>");
  http_write_str(req, "Set the default options");
  http_write_str(req, "</font>");

  http_write_str(req, "<br><br>");


  http_write_str(req, "<table>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_language: ");
  http_write_str(req, "</td><td>\n");

  http_write_str(req, "<SELECT NAME=\"gg_default_language\">");
  http_write_str(req, "<option VALUE=\"\"");
  http_write_str(req, " selected");
  http_write_str(req, ">");
  http_write_str(req, "");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"sq\"");
  http_write_str(req, ">");
  http_write_str(req, "Albanian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ar\"");
  http_write_str(req, ">");
  http_write_str(req, "Arabic");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"bg\"");
  http_write_str(req, ">");
  http_write_str(req, "Bulgarian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ca\"");
  http_write_str(req, ">");
  http_write_str(req, "Catalan");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"zh-CN\"");
  http_write_str(req, ">");
  http_write_str(req, "Chinese (Simplified)");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"zh-TW\"");
  http_write_str(req, ">");
  http_write_str(req, "Chinese (Traditional)");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"hr\"");
  http_write_str(req, ">");
  http_write_str(req, "Croatian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"cs\"");
  http_write_str(req, ">");
  http_write_str(req, "Czech");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"da\"");
  http_write_str(req, ">");
  http_write_str(req, "Danish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"nl\"");
  http_write_str(req, ">");
  http_write_str(req, "Dutch");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"en\"");
  http_write_str(req, ">");
  http_write_str(req, "English");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"et\"");
  http_write_str(req, ">");
  http_write_str(req, "Estonian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"tl\"");
  http_write_str(req, ">");
  http_write_str(req, "Filipino");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"fi\"");
  http_write_str(req, ">");
  http_write_str(req, "Finnish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"fr\"");
  http_write_str(req, ">");
  http_write_str(req, "French");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"gl\"");
  http_write_str(req, ">");
  http_write_str(req, "Galacian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"de\"");
  http_write_str(req, ">");
  http_write_str(req, "German");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"el\"");
  http_write_str(req, ">");
  http_write_str(req, "Greek");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"iw\"");
  http_write_str(req, ">");
  http_write_str(req, "Hebrew");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"hi\"");
  http_write_str(req, ">");
  http_write_str(req, "Hindi");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"hu\"");
  http_write_str(req, ">");
  http_write_str(req, "Hungarian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"id\"");
  http_write_str(req, ">");
  http_write_str(req, "Indonesian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"it\"");
  http_write_str(req, ">");
  http_write_str(req, "Italian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ja\"");
  http_write_str(req, ">");
  http_write_str(req, "Japanese");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ko\"");
  http_write_str(req, ">");
  http_write_str(req, "Korean");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"lv\"");
  http_write_str(req, ">");
  http_write_str(req, "Latvian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"lt\"");
  http_write_str(req, ">");
  http_write_str(req, "Lithuanian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"mt\"");
  http_write_str(req, ">");
  http_write_str(req, "Maltese");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"no\"");
  http_write_str(req, ">");
  http_write_str(req, "Norwegian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"fa\"");
  http_write_str(req, ">");
  http_write_str(req, "Persian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"pl\"");
  http_write_str(req, ">");
  http_write_str(req, "Polish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"pt\"");
  http_write_str(req, ">");
  http_write_str(req, "Portuguese");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ro\"");
  http_write_str(req, ">");
  http_write_str(req, "Romanian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ru\"");
  http_write_str(req, ">");
  http_write_str(req, "Russian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"sr\"");
  http_write_str(req, ">");
  http_write_str(req, "Serbian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"sk\"");
  http_write_str(req, ">");
  http_write_str(req, "Slovak");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"sl\"");
  http_write_str(req, ">");
  http_write_str(req, "Slovenian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"es\"");
  http_write_str(req, ">");
  http_write_str(req, "Spanish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"sv\"");
  http_write_str(req, ">");
  http_write_str(req, "Swedish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"th\"");
  http_write_str(req, ">");
  http_write_str(req, "Thai");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"tr\"");
  http_write_str(req, ">");
  http_write_str(req, "Turkish");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"uk\"");
  http_write_str(req, ">");
  http_write_str(req, "Ukranian");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"vi\"");
  http_write_str(req, ">");
  http_write_str(req, "Vietnamese");
  http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");

  http_write_str(req, "</td><td>");
/*
  http_write_str(req, "<input type=\"text\" NAME=\"gg_default_language\" SIZE=10>&nbsp;");
*/
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_default_language") );
  http_write_str(req, "</td></tr>");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_planet: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_default_planet\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_default_planet") );
  http_write_str(req, "</td>\n");
  http_write_str(req, "</tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_countrycode: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_default_countrycode\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
    http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_default_countrycode") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_postcode: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_default_postcode\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
    http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_default_postcode") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_topic1: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_def_top1\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
    http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_def_top1") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_topic2: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_def_top2\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
    http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_def_top2") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_topic3: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_def_top3\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
    http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_def_top3") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_topic4: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_def_top4\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_def_top4") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_default_topic5: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_def_top5\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_def_top5") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_threshold: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_threshold\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_threshold") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "gg_dl_timeout: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"gg_dl_timeout\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("gg_dl_timeout") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "<tr><td>\n");
  http_write_str(req, "max_results: ");
  http_write_str(req, "</td><td>\n");
  http_write_str(req, "<input type=\"text\" NAME=\"max_results\" SIZE=17>&nbsp;");
  http_write_str(req, "</td><td>\n");
  http_write_va(req, BUFSZ1, " current = [ %s ]", conf_get_str("max_results") );
  http_write_str(req, "</td></tr>\n");

  http_write_str(req, "</table>\n");

  http_write_str(req, "<br>");
  http_write_str(req, "<br>");

  http_write_str(req, "connect to  IP address:[port] : ");
  http_write_str(req, "<input type=\"text\" NAME=\"connect_ip\" SIZE=14>&nbsp;");

  http_write_str(req, "<br>");
  http_write_str(req, "<br>");

  http_write_str(req, "<input value=\"");
  http_write_str(req, "Save Options");
  http_write_str(req, "\" type=\"submit\"><br>");

  http_write_str(req, "<br>");
  http_write_str(req, "<br>");
  http_write_str(req, "<h2>Connection Info :</h2> ");



/****  start . print info */

  /* info_hosts */

  uint h_num;
  float64 bytes, files;
  char buf[256];
  char *t1,*t2;

  host_totals(&h_num, &files, &bytes);

  if (gc_munge) {
    t1 = format_si(files, 0);
    t2 = format_si(bytes, 0);
  } else {
    t1 = ymaloc(20, 356);
    t2 = ymaloc(20, 357);
    sprintf(t1, "%g", files);
    sprintf(t2, "%g M", bytes / 1000000.0);
  }
  sprintf(buf, "HOST STATS:  Hosts: %-6i Files: %-7s Size: %-7s\n",
          h_num, t1, t2);
  fre_str(&t1, 204);
  fre_str(&t2, 205);
  http_write_va(req, BUFSZ1, " %s ", buf );

  http_write_str(req, "<br> ");

  sprintf(buf, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;   LRU Cache: %4u\n", hlru_num);

  http_write_va(req, BUFSZ1, " %s ", buf  );


  /* info_net */
  http_write_str(req, "<br> ");

  int messages, sent;
  float64 n_rcv_bytes, n_snt_bytes;
  float64 n_blacklisted;

  gnut_connection_net_totals(&messages, &sent, &n_rcv_bytes, &n_snt_bytes, &n_blacklisted);
  if (gc_munge) {
    t1 = format_si((float64) messages, 0);
    t2 = format_si((float64) sent, 0);
  } else {
    t1 = ymaloc(20, 352);
    t2 = ymaloc(20, 353);
    sprintf(t1, "%i", messages);
    sprintf(t2, "%i", sent);
  }
  sprintf(buf, UI_INFO_NET, t1, t2);
  fre_str(&t1, 200);
  fre_str(&t2, 201);
  http_write_va(req, BUFSZ1, " %s ", buf  );

  http_write_str(req, "<br> ");

//  sprintf(buf, "             Unique GUIDs in memory: %i\n", route_queue->size);
//  http_write_va(req, BUFSZ1, " %s ", buf  );

  if (gc_munge) {
    t1 = format_si(n_rcv_bytes, 0);
    t2 = format_si(n_snt_bytes, 0);
  } else {
    t1 = ymaloc(20, 354);
    t2 = ymaloc(20, 355);
    sprintf(t1, "%g", n_rcv_bytes);
    sprintf(t2, "%g", n_snt_bytes);
  }
  sprintf(buf, " &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; Bytes Rcvd: %-7s       Bytes Sent: %-7s\n", t1, t2);
  fre_str(&t1, 202);
  fre_str(&t2, 203);
    http_write_va(req, BUFSZ1, " %s ", buf  );


  /*  info_query */

  http_write_str(req, "<br> ");

  sprintf(buf,"QUERY STATS: Queries: %-6u   Responses Sent: %-6u\n",
          con_num_queries, con_num_responses);
  http_write_va(req, BUFSZ1, " %s ", buf  );

#ifdef USE_DRR_SEARCH
  if (t_counter) {
    r = (((long)t_hits) * 1000L) / ((long) t_counter);
  } else {
    r = 0;
  }
  sprintf(buf,"             QREPLY Cache hit rate: %2i.%1i%%\n",
          r/10, r%10);
  http_write_va(req, BUFSZ1, " %s ", buf  );
#endif


  /* info_shares */

  http_write_str(req, "<br> ");

  uint32 num;
  float64 size;

  share_totals(&num, &size);
  if (gc_munge) {
    t1 = format_si(size, 0);
  } else {
    t1 = ymaloc(20, 351);
    sprintf(t1, "%g k bytes", size / 1000.0);
  }
  sprintf(buf, UI_INFO_SHARE, num, t1);
  fre_str(&t1, 199);
  http_write_va(req, BUFSZ1, " %s ", buf  );

  sprintf(buf, urate_clipped ? UI_INFO_SHAR2 : UI_INFO_SHAR3, (best_urate + 64) >> 7, (num_uploads < gc_max_uploads) ? 1 : 0, gh_did_upload, gh_did_receive);

  http_write_str(req, "<br> ");
  http_write_va(req, BUFSZ1, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp; %s ", buf  );


  /* info_connection_callback */

#define FMT_TERSE 1
  int connection_num = 0;

  int info_connection_callback(gcs *gc)
  {
    long total_sent=0;
    float inb=0;
    float outb=0;
    float64 s, r;
    time_t i;
    char *t1,*t2;
    char buf[4096];
    char buf2[100];
    char ipstr[32];
    int fmt;

    fmt = conf_get_int("stats_format");

    connection_num++;
    gc->uindex = connection_num;

    sprintf(ipstr, "%i.%i.%i.%i:%i", gc->ip.b[0], gc->ip.b[1],
            gc->ip.b[2], gc->ip.b[3], gc->port);

    if (gc_munge) {
      t1 = format_si(gc->snt_bytes, 0);
      t2 = format_si(gc->rcv_bytes, 0);
    } else {
      t1 = ymaloc(20, 347);
      sprintf(t1, "%7g", gc->snt_bytes);
      t2 = ymaloc(20, 348);
      sprintf(t2, "%-7g", gc->rcv_bytes);
    }

    http_write_str(req, "<br> ");
    /* First line of stats */
    if (fmt & FMT_TERSE) {
      sprintf(buf,"%2i) %-20s",
              connection_num, ipstr);
      sprintf(buf2, "%i,%i,%i (%2li) B:%i",
              gc->sent_packets, gc->received_packets,
              gc->dropped_packets,
              droprate(gc->dropped_packets, gc->received_packets),
              gc->bad_packets );
    } else {
      sprintf(buf, UI_INFO_CONN1,
              connection_num, ipstr, gc->sent_packets,
              gc->received_packets, gc->dropped_packets, gc->bad_packets,
              t1, t2);
    }
    fre_str(&t1, 193);
    fre_str(&t2, 194);
    http_write_va(req, BUFSZ1, " %s ", buf  );

    total_sent += gc->sent_packets;

    i = time(0) - gc->rate_start_time;
    if (i) {
      s = gc->rate_snt_bytes / ((float64) i);
      t1 = format_si(s, 0);

      r = gc->rate_rcv_bytes / ((float64) i);
      t2 = format_si(r, 0);
    } else {
      t1 = ymaloc(10, 349);
      t2 = ymaloc(10, 350);
      strcpy(t1, "0");
      strcpy(t2, "0");
      s = 0;
      r = 0;
    }
    if (gc->cstate == STATE_CONNECTED) {
      inb += r;
      outb += s;
    }

    /* Second line of stats */
    if (fmt & FMT_TERSE) {
      if (gc->cstate == STATE_CONNECTED) {
        sprintf(buf," %s OK: %s\n", info_str_type(gc->ctype), buf2);
      } else {
        sprintf(buf," %s %s\n", info_str_type(gc->ctype),
                info_str_state(gc->cstate, 4));
      }
    } else {
      sprintf(buf, UI_INFO_CONN2,
#if defined(hpux) && defined(PTHREADS_DRAFT4)
              ((uint32) ((HPUX_hack) gc->tid).longpart), /* 0.4.28.c12 */
#else
              (uint32) gc->tid,
#endif
              info_str_type(gc->ctype), info_str_state(gc->cstate, 4),
              t1, t2);
    }

    fre_str(&t1, 195);
    fre_str(&t2, 196);
    http_write_va(req, BUFSZ1, " %s ", buf  );

  #ifdef MOREPACKETSTATS
    if (conf_get_int("packet_stats")) {
      int n,n2;

      for (n = 0; n < 256; n++) {
        if (gc->packet_count[n]) {
          sprintf(buf, UI_INFO_CONN3,
                  n, gc->packet_count[n],
                  gc->packet_byte_count[n]/gc->packet_count[n],
                  gc->max_packet_size[n]);
          http_write_va(req, BUFSZ1, " %s ", buf  );
        }
      }

      if (gc->packet_count[0]) {
        n2 = sprintf(buf, UI_INFO_CONN4);
        for (n = 0; n < 256; n++) {
          if (gc->ttl_count[n]) {
            n2 += sprintf(&buf[n2], "%d(%d) ", n, gc->ttl_count[n]);
          }
        }
        sprintf(&buf[n2],"\n");
        http_write_va(req, BUFSZ1, " %s ", buf  );
      }

      for (n = 0; n < 256; n++) {
        if (gc->out_packet_count[n]) {
          sprintf(buf, UI_INFO_CONN5, n, gc->out_packet_count[n],
                  gc->out_packet_byte_count[n]/gc->out_packet_count[n],
                  gc->out_max_packet_size[n]);
          http_write_va(req, BUFSZ1, " %s ", buf  );
        }
      }

      if (gc->out_packet_count[0]) {
        n2 = sprintf(buf, UI_INFO_CONN6);
        for (n = 0; n < 256; n++) {
          if (gc->out_ttl_count[n]) {
            n2 += sprintf(&buf[n2],"%d(%d) ",
                          n, gc->out_ttl_count[n]);
          }
        }
        sprintf(&buf[n2],"\n");
        http_write_va(req, BUFSZ1, " %s ", buf  );
      }
    }
  #endif

    return 0;
  }


  /* info_connection */

  float inb=0;
  float outb=0;
  long total_sent;


  if (! gnut_connection_num()==0) {

    http_write_str(req, "<br> ");
    http_write_str(req, "<br> ");
    http_write_str(req, "<br> ");
    http_write_str(req, "CONNECTION STATS: ");
    http_write_str(req, "<br> ");

    http_write_str(req, "----------------- ");
    connection_num = 0;
    inb = 0;
    outb = 0;
    total_sent = 0;

    gnut_connection_enumerate(info_connection_callback);

    t1 = format_si((unsigned long) outb, 0);
    t2 = format_si((unsigned long) inb, 0);
    sprintf(buf, UI_INFO_CONN_TOTALS, t1, t2, total_sent);
    fre_str(&t1, 197);
    fre_str(&t2, 198);
    http_write_str(req, "<br> ");
    http_write_va(req, BUFSZ1, " %s ", buf  );
    http_write_str(req, "<br><br> ");

  }


/****  end . print info */



  http_write_str(req, "</FORM>");

  return 0;
}

int gnut_http_write_form_create(http_request_t *req, int ttw, int rets)
{

  GD_S(2,"gnut_http_write_form_create entering\n");
  http_write_str(req, "<FORM ACTION=\"/create/\">");

  http_write_str(req, "<font size=+3>");
  http_write_str(req, "Create a new proposal");
  http_write_str(req, "</font><br><br>");


  /// timestamps  ///
  http_write_str(req, "Time Stamp:");
  http_write_str(req, "<input type=\"text\" NAME=\"timestamps\" SIZE=10>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"time_stamps\">");
  http_write_str(req, "<option VALUE=\"\"");
  http_write_str(req, " selected");
  http_write_str(req, ">");
  http_write_str(req, "");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"Current Time\"");
  http_write_str(req, ">");
  http_write_str(req, "Current Time");
  http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");

  http_write_str(req, "<br>");
  http_write_str(req, "<br>");

  /// country codes  ///
  http_write_str(req, "Country code:");
  http_write_str(req, "<input type=\"text\" NAME=\"countrycodes\" SIZE=3>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"recent_cc\">");
  http_write_str(req, "<option VALUE=\"\"");
  http_write_str(req, " selected");
  http_write_str(req, ">");
  http_write_str(req, "");
  http_write_str(req, "</option>");

  http_write_va(req, BUFSZ1, "<option VALUE=\"%s\"", conf_get_str("gg_default_countrycode"));
  http_write_str(req, ">");
  http_write_va(req, BUFSZ1, "%s", conf_get_str("gg_default_countrycode"));
  http_write_str(req, "</option>");

  http_write_str(req, "<option VALUE=\"au\"");
  http_write_str(req, ">");
  http_write_str(req, "au");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"uk\"");
  http_write_str(req, ">");
  http_write_str(req, "uk");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"ru\"");
  http_write_str(req, ">");
  http_write_str(req, "ru");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"my\"");
  http_write_str(req, ">");
  http_write_str(req, "my");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"us\"");
  http_write_str(req, ">");
  http_write_str(req, "us");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"in\"");
  http_write_str(req, ">");
  http_write_str(req, "in");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"cn\"");
  http_write_str(req, ">");
  http_write_str(req, "cn");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"es\"");
  http_write_str(req, ">");
  http_write_str(req, "es");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"nz\"");
  http_write_str(req, ">");
  http_write_str(req, "nz");
  http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");
  http_write_str(req, "&nbsp &nbsp &nbsp &nbsp &nbsp &nbsp Postcode:");
  http_write_str(req, "<input type=\"text\" NAME=\"postcodes\" SIZE=5>&nbsp;");

  http_write_str(req, "<SELECT NAME=\"recent_pc\">");
  http_write_str(req, "<option VALUE=\"\"");
  http_write_str(req, " selected");
  http_write_str(req, ">");
  http_write_str(req, "");
  http_write_str(req, "</option>");
  http_write_va(req, BUFSZ1, "<option VALUE=\"%s\"", conf_get_str("gg_default_postcode"));
  http_write_str(req, ">");
  http_write_va(req, BUFSZ1, "%s", conf_get_str("gg_default_postcode"));
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"6000\"");
  http_write_str(req, ">");
  http_write_str(req, "6000");
  http_write_str(req, "</option>");
  http_write_str(req, "<option VALUE=\"6157\"");
  http_write_str(req, ">");
  http_write_str(req, "6157");
  http_write_str(req, "</option>");

  http_write_str(req, "</SELECT>");
  http_write_str(req, "<br>");
  http_write_str(req, "<br>");

  http_write_str(req, "Topic1:");
  http_write_str(req, "<input type=\"text\" NAME=\"topic1\" SIZE=60>&nbsp;");
  http_write_str(req, "<br>");
  http_write_str(req, "Topic2:");
  http_write_str(req, "<input type=\"text\" NAME=\"topic2\" SIZE=60>&nbsp;");
  http_write_str(req, "<br>");
  http_write_str(req, "Topic3:");
  http_write_str(req, "<input type=\"text\" NAME=\"topic3\" SIZE=60>&nbsp;");
  http_write_str(req, "<br>");
  http_write_str(req, "Topic4:");
  http_write_str(req, "<input type=\"text\" NAME=\"topic4\" SIZE=60>&nbsp;");
  http_write_str(req, "<br>");
  http_write_str(req, "Topic5:");
  http_write_str(req, "<input type=\"text\" NAME=\"topic5\" SIZE=60>&nbsp;");
  http_write_str(req, "<br>");
  http_write_str(req, "<br>");

  http_write_str(req, "Contents");
  http_write_str(req, "<br>");
  http_write_str(req, "<textarea NAME=\"content\" cols=\"60\" rows=\"8\"></textarea>");
  http_write_str(req, "<input value=\"");
  http_write_str(req, "Create");
  http_write_str(req, "\" type=\"submit\"><br>");


  time_t tim;
  char ts[256] = "";

  tim = time(0);
  strftime(ts,16,"%Y%m%d%H%M%S",localtime(&tim));

  // http_write_str(req, "<br>");
  // ggggggg 111111  http_write_va(req, BUFSZ1, "time== %d  <br>", tim);
  //  http_write_va(req, BUFSZ1, "ts== %s  <br>", ts);

  http_write_str(req, "</FORM>");

  return 0;
}

void gnut_http_template_scan()
{
  FILE *fp;
  char *fname;
  long len;

  if (!conf_get_str("html_template")) {
    return;
  }
  if (strlen(conf_get_str("html_template"))<=1) {
    return;
  }
  fname = expand_path(conf_get_str("html_template"));
  GD_S(1, "path to html template: "); GD_S(1, fname); GD_S(1, "\n");
  fp=fopen(fname,"r");

  if (fp == 0) {

    printf("Creating html template file \"%s\".\n", fname);

    fp = fopen(fname, "w");

    fprintf(fp, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n");
    fprintf(fp, "<html>\n");
    fprintf(fp, "<head>\n");
    fprintf(fp, "<title>gumnut - Organisational Software</title>\n");
    fprintf(fp, "</head>\n");
    fprintf(fp, "<body bgcolor=\"#FFFFFF\">\n");
    fprintf(fp, "<table width=\"100%%\">\n");
    fprintf(fp, "<tbody>\n");
    fprintf(fp, "<tr>\n");
    fprintf(fp, "<td align=\"center\" bgcolor=\"#3366ff\"><font size=\"+3\" face=\"Helvetica\" color=\"#dddddd\"> gumnut</font></td>\n");
    fprintf(fp, "</tr>\n");
    fprintf(fp, "</tbody>\n");
    fprintf(fp, "</table> \n");
    fprintf(fp, "<br>\n");
    fprintf(fp, "\n");
    fprintf(fp, "<!-- gnut -->\n");
    fprintf(fp, "\n");
    fprintf(fp, "<br>\n");
    fprintf(fp, "gumnut client file server and search gateway Powered by gnut v0.4.28<br>\n");
    fprintf(fp, "Related web sites:<a href= \"http://gumnut.sf.net/\">gumnut.sf.net</a><br>\n");
    fprintf(fp, "</body>\n");
    fprintf(fp, "</html>\n");
    fprintf(fp, "\n");

    fclose(fp);

  }


  fp=fopen(fname,"r");

  if (fp == 0) {
    GD_S(1, "Couldn't open html_template "); GD_S(1, fname); GD_S(1, "\n");
    fre_str(&fname, 181);
  } else {
    fre_str(&fname, 182);

    /*  find the length */
    fseek(fp, 0, SEEK_END);
    len = ftell(fp);
    rewind(fp);
    GD_S(3, "html_template length: "); GD_I(3, len); GD_S(3, "\n");

    /* real.loc space for the entire file */
    if (html_template_head) {
      fre_str(&html_template_head, 490);
    }
    html_template_head = ymaloc(len+1, 491);
    if (!html_template_head) {
      GD_S(1, "Unable to real""loc space for HTML template.\n");
      return;
    }

    /* slurp! */
    fread(html_template_head, sizeof(char), len, fp);
    fclose(fp);

    /* search for comment */
    html_template_foot=strstr(html_template_head,"<!-- gnut -->");
    if (html_template_foot) {
      *html_template_foot = '\0';
      html_template_foot+=13;
    } else {
      GD_S(1, "Comment not found.\n");
      html_template_foot=&html_template_head[len];
    }
  }
}
#endif

static void gnut_guid_to_string(char *string, char *guid)
{
  int i;
  for (i = 0; i < 16; i++) {
   sprintf(&string[i*2], "%02x", (unsigned char)guid[i]);
  }
}

void fre_hth(http_header_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

void fre_hrt(http_request_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

void fre_hct(http_cookie_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

http_request_t *http_request_new(int sock)
{
  http_request_t *req;

  GD_S(HTTP_REQUEST_DEBUG, "http_request_new IN\n");
  req = ymaloc(sizeof(http_request_t), 414);
  req->method = 0;
  req->version = 0;
  req->url_file = 0;
  req->rq_headers = 0;
  req->response_headers = 0;
  req->sock = sock;
  GD_S(HTTP_REQUEST_DEBUG, "http_request_new OUT\n");
  return req;
}

void http_request_delete(http_request_t *req)
{
  GD_S(HTTP_REQUEST_DEBUG, "http_request_delete IN\n");
  if (req) {
    if (req->method) {
      fre_str(&(req->method), 270);
    }
    if (req->version) {
      fre_str(&(req->version), 271);
    }
    if (req->url_file) {
      fre_str(&(req->url_file), 272);
    }

    /* HACK - slightly leaky - we should
     * walk the list and fr.ee the contents first. */
    if (req->rq_headers) {
      gnut_list_fre(req->rq_headers);
    }
    if (req->response_headers) {
      gnut_list_fre(req->response_headers);
    }
    fre_hrt(&req, 273);
  }
  GD_S(HTTP_REQUEST_DEBUG, "http_request_delete OUT\n");
}

/* Break the HTTP request into its parts. */
int http_request_parse(http_request_t *req, char *request)
{
  char *ptr_to_http;
  char *ptr_to_file;
  int len;

  GD_S(HTTP_REQUEST_DEBUG, "http_request_parse IN\n");

  ptr_to_file = strstr(request, " ");
  if (!ptr_to_file) {
    GD_S(HTTP_REQUEST_DEBUG, "http_request_parse OUT - nothing to parse\n");
    return -1;
  }
  ptr_to_file++;

  ptr_to_http=strstr(request,"HTTP/");
  if (ptr_to_http == 0) {
    GD_S(HTTP_REQUEST_DEBUG, "http_request_parse OUT - not an HTTP/1.0 or higher request\n");
    return -1;
  }

  /*
   * Save the "method" portion of the request.
   */
  len = (ptr_to_file - request);
  req->method = ymaloc(sizeof(char)*(len+1), 415);
  strncpy(req->method, request, len);
  *(req->method+len) = '\0';

  /*
   * Save the "url" portion of the request.
   */
  len = (ptr_to_http-1-ptr_to_file);
  req->url_file = ymaloc(sizeof(char)*(len+1), 416);
  strncpy(req->url_file, ptr_to_file, len);
  *(req->url_file+len) = '\0';

  /*
   * Save the HTTP version of the request - just in case we need it.
   */
  len = strlen(ptr_to_http);
  req->version = ymaloc(sizeof(char)*(len+1), 417);
  strncpy(req->version, ptr_to_http, len);
  *(req->version + len) = '\0';

  GD_S(HTTP_REQUEST_DEBUG, "http_request_parse OUT\n");

  return 0;
}

/* Examine the HTTP headers and find one of the values. */
char *http_request_find_header(http_request_t *req, char *name)
{
  Gnut_List *list;
  http_header_t *hdr;

  GD_S(HTTP_REQUEST_DEBUG, "http_request_find_header IN - ");
  GD_S(HTTP_REQUEST_DEBUG, name);
  GD_S(HTTP_REQUEST_DEBUG, "\n");

  list = req->rq_headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;

    if (!strcmp(hdr->name, name)) {
      GD_S(HTTP_REQUEST_DEBUG, "http_request_find_header OUT - found header ");
      GD_S(HTTP_REQUEST_DEBUG, hdr->value);
      GD_S(HTTP_REQUEST_DEBUG, "\n");
      return hdr->value;
    }

    list = gnut_list_next(list);
  }

  GD_S(HTTP_REQUEST_DEBUG, "http_request_find_header OUT - no such header\n");

  return 0;
}

/* Read HTTP headers into the request object. */
int http_request_read_header(http_request_t *req)
{
  char buf[4096];
  int ret;
  http_header_t *hdr;

  GD_S(HTTP_REQUEST_DEBUG, "http_request_read_header IN\n");

  /* We must first read the remaining headers
   * and ignore them. */
  ret = 1;
  while (1) {
    errno = 0;
    ret = read_line(req->sock, buf, 4096);
    if (gc_debug_opts & 4) {
      printf("hrh01 %s", buf);
    }
    if (errno == EINTR )
      break;
    if (strlen(buf) < 3)
      break;

    hdr = http_header_new();
    if (!hdr)
      break;

    ret = http_header_parse(hdr, buf);
    if (ret) {
      break;
    }

    GD_S(1, "Got header "); GD_S(1, hdr->name); GD_S(1, ":");
    GD_S(1, hdr->value); GD_S(1, "\n");

    req->rq_headers = gnut_list_prepend(req->rq_headers, hdr);
  }
  GD_S(HTTP_REQUEST_DEBUG, "http_request_read_header OUT\n");
  return 0;
}

/* Set the contents of an HTTP header. */
int http_response_header_add(http_request_t *req, char *name, char *value)
{
  http_header_t *hdr;

  GD_S(HTTP_HEADER_DEBUG, "http_response_header_add IN - ");
  GD_S(HTTP_HEADER_DEBUG, name);
  GD_S(HTTP_HEADER_DEBUG, " - ");
  GD_S(HTTP_HEADER_DEBUG, value);
  GD_S(HTTP_HEADER_DEBUG, "\n");

  hdr = http_header_new();
  http_header_set(hdr, name, value);
  req->response_headers = gnut_list_prepend(req->response_headers, hdr);

  GD_S(HTTP_HEADER_DEBUG, "http_response_header_add OUT\n");
  return 0;
}

int http_response_header_add_va(http_request_t *req, char *name, char *format, ...)
{
  va_list argp;
  char hdr_value[64];
  int ret;

  GD_S(HTTP_HEADER_DEBUG, "http_response_header_add_va IN\n");

  va_start(argp,format);
  vsprintf(hdr_value,format,argp);
  va_end(argp);
  ret = http_response_header_add(req, name, hdr_value);

  GD_S(HTTP_HEADER_DEBUG, "http_response_header_add_va OUT\n");
  return ret;
}

http_cookie_t *http_cookie_new(void)
{
  http_cookie_t *cookp;

  cookp = ymaloc(sizeof(http_cookie_t), 418);
  cookp->value = 0;
  cookp->path = 0;
  cookp->domain = 0;
  cookp->expiration = time(0);

  return cookp;
}

int http_header_get_cookie(http_request_t *req, http_cookie_t *cookp)
{
  char *end_of_value;
  int len_of_value;
  char *header_value;

  header_value = http_request_find_header(req, "Cookie");
  if (!header_value) return -1;

  end_of_value = strchr(header_value, ';');
  if (!end_of_value) {
    len_of_value = strlen(header_value)+1;
  }
  else {
    len_of_value = end_of_value - header_value;
  }

  cookp->value = ymaloc(sizeof(char)*len_of_value, 419);
  if (!cookp->value) {
    return -1;
  }
  strncpy(cookp->value, header_value, len_of_value);
  cookp->value[len_of_value-1] = '\0';

  return 0;
}

void http_cookie_delete(http_cookie_t *cookie)
{
  if (cookie) {
    if (cookie->value) {
      fre_str(&(cookie->value), 274);
    }
    if (cookie->path) {
      fre_str(&(cookie->path), 275);
    }
    if (cookie->domain) {
      fre_str(&(cookie->domain), 276);
    }
    fre_hct(&cookie, 558);
  }
}

int http_response_set_cookie(http_request_t *req, http_cookie_t *cookie)
{
  return http_response_header_add_va(req, "Set-Cookie", "%s; path=%s", cookie->value, "/");
}

int http_response_header_write(http_request_t *req)
{
  http_header_t *hdr;
  Gnut_List *list;

  GD_S(HTTP_REQUEST_DEBUG, "http_response_header_write IN\n");

  /* If returning a partial file, we would want to write
   * "206 Partial Content", but alas, the original Gnutella
   * (Nullsoft Windows client) doesn't work if we do that */
  http_write_str(req, "HTTP/1.0 200 OK\r\n"); /* tagok */

  list = req->response_headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;
    http_write_va(req, HTTP_MAX_HEADER_SIZE, "%s: %s\r\n", hdr->name, hdr->value);
    list = gnut_list_next(list);
  }

  http_write_str(req, "\r\n");

  GD_S(HTTP_REQUEST_DEBUG, "http_response_header_write OUT\n");

  return 0;
}

/* Create a new header object. */
http_header_t *http_header_new(void)
{
  http_header_t *hdr;

  GD_S(HTTP_HEADER_DEBUG, "http_header_new IN\n");

  hdr = (http_header_t *)ymaloc(sizeof(http_header_t), 307);
  if (hdr) {
    hdr->name = 0;
    hdr->value = 0;
  }

  GD_S(HTTP_HEADER_DEBUG, "http_header_new OUT\n");

  return hdr;
}

/* Parse an HTTP header from a header line. */
int http_header_parse(http_header_t *hdr, char *hdr_line)
{
  char *value;
  char *s;
  int len;

  GD_S(HTTP_HEADER_DEBUG, "http_header_parse IN\n");

  value = strchr(hdr_line, ':');
  if (!value) {
    GD_S(HTTP_HEADER_DEBUG, "This is not a proper header\n");
    return -1;
  }

  len = value - hdr_line;
  hdr->name = ymaloc(sizeof(char)*(len+1), 308);
  if (hdr->name) {
    strncpy(hdr->name, hdr_line, len);
    *(hdr->name+len) = '\0';
  }

  /* Skip past the ":" and the space. */
  value++;
  if (*value) {
    value++;
  }

  if (!(*value)) {
    if (hdr->name) {
      fre_str(&(hdr->name), 380);
    }
    fre_hth(&hdr, 381);
    GD_S(HTTP_HEADER_DEBUG, "http_header_parse OUT - not a proper header\n");
    return -1;
  }

  len = strlen(value)-2;
  hdr->value = ymaloc(sizeof(char)*(len+1), 309);
  if (hdr->value) {
    strncpy(hdr->value, value, len);
    *(hdr->value+len) = '\0';

    s = hdr->value+(len-1);
    while (G_ISSPACE(*s)) {
      *s = '\0';
    }
  }

  GD_S(HTTP_HEADER_DEBUG, "http_header_parse OUT\n");

  return 0;
}

/* Set the contents of an HTTP header. */
int http_header_set(http_header_t *hdr, char *name, char *value)
{
  GD_S(HTTP_HEADER_DEBUG, "http_header_set IN\n");
  hdr->name = ystdup(name, 455);
  hdr->value = ystdup(value, 456);
  GD_S(HTTP_HEADER_DEBUG, "http_header_set OUT\n");
  return 0;
}

void http_header_delete(http_header_t *hdr)
{
  GD_S(HTTP_HEADER_DEBUG, "http_header_delete IN\n");
  if (hdr) {
    if (hdr->name) {
      fre_str(&(hdr->name), 376);
    }
    if (hdr->value) {
      fre_str(&(hdr->value), 377);
    }
    fre_hth(&hdr, 378);
  }
  GD_S(HTTP_HEADER_DEBUG, "http_header_delete OUT\n");
}

#if 0

 these are sent by clients getting files from us

 User-Agent: Java1.1.8
 Host: 24.128.226.169:5634
 Accept: text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2              */
 Connection: keep-alive

 User-Agent: Gnutella

 User-Agent: Gnotella
 Range: bytes=0-

 User-Agent: Gnotella 0.9.9
 Referer: Gnutella
 Connection: Keep-Alive

 Connection: Keep-Alive
 Range: bytes=0-

 Host: 24.128.226.169:5634
 Accept: */*                                                               */
 Range: bytes=1-1
 User-Agent: RealDownload/4.0.0.41

 User-Agent: LimeWire
 Range: bytes=0-

 User-Agent: LimeWire 1.4
 Range: bytes=0-

 User-Agent: BearShare 1.3.1
 Connection: Keep-Alive
 Range: bytes=0-

 User-Agent: BearShare 1.3.2
 Connection: Keep-Alive
 Range: bytes=0-

 User-Agent: BearShare 2.0.0
 Referrer: Gnutella file sharing network
 Connection: Keep-Alive
 Range: bytes=0-

 Server: BearShare 2.0.1
 Content-type:application/binary
 Content-Length:3616520
 Content-Range: bytes=0-3616519/3616520

 User-Agent: BearShare 2.0.4
 Referrer: Gnutella
 Connection: Keep-Alive
 Range: bytes 0-

 User-Agent: BearShare 2.0.9
 Referrer: Gnutella
 Connection: Keep-Alive
 Range: bytes 0-

 User-Agent: BearShare 2.2.3
 Referrer: Gnutella
 Connection: Keep-Alive
 Range: bytes=0-

 these are returned by servers when we do downloads

 HTTP 200 OK                                tagok
 Server: Gnutella
 Content-type:application/binary
 Content-length:3071634

 HTTP 200 OK                                tagok
 Server: BearShare 2.0.0
 Content-type:application/binary
 Content-Length:5016346
 Content-Range: bytes=0-5016345/5016346

 HTTP 503 Duplicate Request                 tagok
 Server: BearShare 2.0.0

 HTTP 503 Server Busy                       tagok
 Server: BearShare 2.0.1

#endif
